"""
	For use in Ignition's Jython environment.

	Adds tags for the simulator's variables
"""


import re
from java.util import Date



class TagsMixin(object):

	_TAG_TYPE_MAP = {
		str: 'String',
		int: 'Int4',
		float: 'Float8',
		bool: 'Boolean',
		Date: 'DateTime',
		}

	#https://regex101.com/r/ogqErX/4
	_TAG_PATTERN = re.compile(r"""
		^ (?P<fullpath>
		  # If a tag provider is given, match that first
		  (\[(?P<provider>[a-z0-9_\- ]+)\])?
		  # Capture the full path as well
		  (?P<path>
		  # Everything after the provider is a parent path
		  # After the parent path, no forward slashes can be
		  #   matched, so the negative lookahead assures that.
		  ((?P<parent>.*?)[/\\])?(?![/\\])
		  # The tag name won't have a forward slash, and if
		  #   missing we'll check that it's a folder
		  (?P<name>[a-z0-9_\- ]+)?)
		) $
		""", re.X + re.I)


	def __init__(self, tag_path=None, **configuration):
		
		self._tag_folder = tag_path
		
		super(TagsMixin, self).__init__(**configuration)

		self._initialize_tags()
	
	
	def _initialize_tags(self):
		root_parts = {
				'provider': 'default',
				'parent': ''
			}
			
		if system.tag.exists(self._tag_folder):
			system.tag.removeTag(self._tag_folder)

		root_parts.update(
			dict((k,v) 
			     for k,v 
			     in self._TAG_PATTERN.match(self._tag_folder).groupdict().items() 
			     if v))
				
		system.tag.addTag(
				parentPath='[%(provider)s]%(parent)s' % root_parts,
				name=root_parts['name'],
				tagType='Folder',
			)
		
		self._tag_folder = '[%(provider)s]%(parent)s/%(name)s' % root_parts

		if self._raw_definition:
			system.tag.addTag(
					parentPath=self._tag_folder,
					name='_definition_',
					tagType='MEMORY',
					dataType='String',
					value=self._raw_definition,
				)
		

		system.tag.addTag(
				parentPath=self._tag_folder,
				name='state',
				tagType='MEMORY',
				dataType='String',
				value=self.state,
			)
	
		for variable, value in self._variables.items():
			system.tag.addTag(
				parentPath=self._tag_folder,
				name=variable,
				tagType='MEMORY',
				dataType=self._TAG_TYPE_MAP[type(value)],
				value=value,
			)
	
	
	def step(self):
		super(TagsMixin, self).step()
		
		tag_paths = ['%s/%s' % (self._tag_folder, variable) for variable in sorted(self._variables)]
		values = [value for _,value in sorted(self._variables.items())]
		
		tag_paths.append('%s/state' % self._tag_folder)
		values.append(self.state)
		
		system.tag.writeAll(tag_paths, values)



from .chunk import ChunkScanner


class GroupScanner(ChunkScanner):
    """Returns the whole group when emitting."""
    def __iter__(self):
        self._pending_finally()
        for group in self.source._groups[self._group_cursor:]:
            self._group_cursor += 1
            yield group



import unittest


from ligature.recordset import RecordSet
from ligature.examples import simpleRecordSet, simpleAddition

from ligature.calculations.cluster import Cluster


class ClusterTestCase(unittest.TestCase):

	def test_basic(self):

		function = lambda a,b: a+b

		srs = RecordSet(simpleRecordSet)
		
		c = Cluster([srs], function, 'c')

		# Calculations are lazily evaluated
		self.assertEqual(c._resultSet._groups, [])

		# When evaluated, we get the following
		self.assertEqual(
			[[v.c for v in group] for group in c.results.groups],
			[[1, 3, 3, 5], [5, 7], [7, 9, 9]]
			)

		srs.extend(simpleAddition)

		# adding data from a source doesn't immediately update
		self.assertEqual(len(c._resultSet._groups), 3)

		# but upon evaluation we see an update has been applied
		# note that this maintains groups
		self.assertEqual(
			[[v.c for v in group] for group in c.results.groups],
			[[1, 3, 3, 5], [5, 7], [7, 9, 9], [12, 12, 14], [14, 16, 16]]
			)

		# Demonstrate slicing for columns works as expected
		self.assertEqual(
			[tuple(group) for group in c.results['c',:]],
			[(1, 3, 3, 5), (5, 7), (7, 9, 9), (12, 12, 14), (14, 16, 16)]
			)



def runTests():
	suite = unittest.TestLoader().loadTestsFromTestCase(ClusterTestCase)
	unittest.TextTestRunner(verbosity=2).run(suite)



if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
import math, re
from types import FunctionType

from transitions import State, Machine

try:
    _ = property.setter
except AttributeError:
    from ligature._compat import property

try:
    from yaml import load as yaml_loader, FullLoader
except ImportError:
    from shared.tools.yaml.core import load as yaml_loader, FullLoader


from ligature.simulators.mixins.support import MixinFunctionSupport


class WrappedSimulationFunction(object):
    
    def __init__(self, datasource, function, aliases=None, name_override=None):

        self._datasource = datasource
        self._function = function
                
        if aliases is None:
            self._aliases = {}
        else:
            self._aliases = aliases.copy()
        
        self._resolve()
        
        
    def _resolve(self, name_override=None):
        self._resolve_name(name_override)
        self._resolve_arguments()
        self._resolve_getters()

        
    # force a recalculation if these values are set
    @property
    def function(self):
        return self._function
    
    @function.setter
    def function(self, function):
        if not function == self._function:
            self._function = function
            self._resolve()
    
    
    @property
    def datasource(self):
        return self._datasource
    
    @datasource.setter
    def datasource(self, datasource):
        self._datasource = datasource
        self._resolve()
    
    
    @property
    def aliases(self):
        return self._aliases
    
    @aliases.setter
    def aliases(self, aliases=None):
        if not aliases == self._aliases:
            if aliases is None:
                self._aliases = {}
            else:
                self._aliases = aliases.copy()
            self._resolve()
            
    
    @property
    def arguments(self):
        return self._arguments
    
    def _resolve_arguments(self):
        self._arguments = self._datasource._resolve_arguments(self._function)
        
    
    @property
    def name(self):
        return self._name
    
    def _resolve_name(self, override=None):
        if override:
            self._name = override
        else:
            try: # function???
                self._name = self._function.func_name
            except AttributeError: # ok prolly a class then
                self._name = function_name = self._function.__class__.__name__    
    
    
    
    def current_value(self, variable):
        return self._datasource._variables[variable]
    
    @property
    def source_variables(self):
        return self._datasource._variables.keys()
    
    
    def _resolve_getters(self):
        getters = []
        for argix, argument in enumerate(self.arguments):
            if argument in self.aliases:
                argument = self.aliases[argument]
#             else:
#                 # by default, skip the self
#                 if argix == 0 and argument == 'self':
#                     continue
#             # currently only the Process interface is expected
#             assert argument in self.source_variables, 'Source class should expose current variable values.'
            
            # assume and hope for the best that the arguments left out have defaults
            if not argument in self.source_variables:
                continue
            getters.append(lambda self=self, argument=argument: self.current_value(argument) )
        
        self._getters = getters
        
        
    def __call__(self):
        """Wraps the function call to ensure it gets the most recent values for the bound variables."""
        return self.function(*tuple(arg_func() for arg_func in self._getters))
        
    def __repr__(self):
        return '<Sim-λ %s>' % self.name
        

class Process(MixinFunctionSupport, Machine):
    """
    The core of the simulation design. 
    
    Provide variables and their functions. Set the initial conditions. 
    
    """

    _DEFAULT_START_VALUE = 0
    _DEFAULT_ESCAPEMENT_VARIABLE = 't'
    _TRANSITION_CHECK = 'check_state'
    
    def __init__(self, 
                 # Raw configuration
                 raw_definition=None,
                 # Simulation configuration
                 variables=None, start=None, alias=None, escapement=None,
                 # State machine configuration
                 states=None, transitions=None,
                 # Remaining state machine configuration pass through
                 **keyword_arguments):
        
        # For self reference and reloading
        self._raw_definition = raw_definition
        
        # For configuration
        self._aliases = alias or {}
        self._start_values = start or {}
        self._definitions = dict((variable, lambda: self._DEFAULT_START_VALUE) 
                                 for variable in variables)
        
        # To be initialized
        self._variables = {}
        self._functions = {}
        
        self._escapement_definition = escapement
        self._initialize_escapement()
                
        self._initialize()

        # Machine setup/prep
        self._state_variable_definitions = states
        states = [state for state in states]
        
        self._initialize_transitions(transitions)
        
        keyword_arguments['states'] = states
        keyword_arguments['transitions'] = self._transition_definitions
        keyword_arguments['after_state_change'] = self._reconfigure_after_state_change
        keyword_arguments['auto_transitions'] = False
        super(Process, self).__init__(**keyword_arguments)
        
        # Initial state (first doesn't trigger after state change)
        self._reconfigure_after_state_change()
                
        
    def _initialize_escapement(self):
        """
        Though not required, it is helpful to directly define how the
        clocks ticks forward in a simulation. Some might use an 
        integer increment, follow wall time, follow a fake clock that
        runs faster than normal, etc.
        """
        if self._escapement_definition is None:
            self._escapement_variable = None
            self._escapement = None
            return # nothing to do - assume all functions are parametric
        kind = self._escapement_definition['kind']
        config = self._escapement_definition['config']
        
        self._escapement_variable = config.get('variable', self._DEFAULT_ESCAPEMENT_VARIABLE)
        
        self._definitions[self._escapement_variable] = self._escapement_definition
        
        if kind == 'increment':
            increment = config.get('increment', 1)
            def tick(self=self, increment=increment):
                self._variables[self._escapement_variable] += increment
            self._escapement = tick
            
        else:
            raise NotImplementedError, "Escapement not implemented yet - %s" % kind
        
        self._start_values[self._escapement_variable] = config.get('start', self._DEFAULT_START_VALUE)
        self._start_values['_t_step']  = 0
        self._start_values['_t_state'] = self._start_values[self._escapement_variable] # starts with a state, after all
        
        
    def _initialize(self):
        """
        Initialize the variables to their starting values, if any.
        """
        for variable in self._definitions:
            if not variable in self._start_values:
                self._start_values[variable] = self._DEFAULT_START_VALUE
                
        # prime varaible listing for reference during resolution
        self._variables = self._start_values.copy()
        
        self._variables['_n_states'] = 1
        self._variables['_n_steps'] = 0
        
        self._initialize_variables()
        
        
    def _initialize_variables(self):
        # maintain the status quo if no reconfiguration
        if not self._definitions:
            return
        
        for variable, definition in self._definitions.items():            
            # Do not mutate the variable that controls the stepping, if provided
            if variable == self._escapement_variable:
                continue

            definition = self._resolve_variable_definition(definition)
            
            if not definition or not isinstance(definition, dict):
                continue
                
            # specific variable definitions override more general ones
            if 'default' in definition:
                self._variables[variable] = self._start_values[variable] = definition['default']
                            
            # update definition for variable's aliases for resolution process
            alias = self._aliases.copy()
            alias.update(definition.get('alias', {}))
            definition['alias'] = alias 

            self._functions[variable] = self._resolve_function(variable, definition)
    
    
    def _resolve_variable_definition(self, variable_definition):
        """
        Generate a more complete variable definition. Exists mostly to be overridden.
        (For example, a variable may be string, but should be an Expression instead;
         but why hassle with all the configuration keys when a string will do?)
        """
        if isinstance(variable_definition, dict) and 'kind' in variable_definition:
            return variable_definition
        elif isinstance(variable_definition, FunctionType):
            return dict(kind='', config=dict(function=variable_definition))
        elif variable_definition is None:
            return dict(kind='', config=dict(hold=None))
        else:
            return dict(kind='', config=dict(value=variable_definition))
            
            
    def _resolve_arguments(self, some_callable):
        """
        Allow mixins to define how their arguments are resolved for the function closures.
        Returns a list or tuple of argument names that the callable expects.
        """
        try: # function???
            return some_callable.func_code.co_varnames[:some_callable.func_code.co_argcount]
        except AttributeError: # maybe it's a callable class thing...
            try:
                return some_callable.__call__.func_code.co_varnames[:some_callable.__call__.func_code.co_argcount]    
            except AttributeError:
                return tuple() # if the buck gets passed up the chain all the way to here, well, /shrug
        #return super(FunctionMixin, self)._resolve_arguments(some_callable)
        
                
    def _resolve_function(self, variable, definition):
        """
        Given a definition, generate a function for a variable to follow.
        """
        assert isinstance(definition, dict), 'Variable function definition for "%s" should be a dictionary.' % variable
        assert set(('kind','config',)).intersection(set(definition.keys())), 'Variable definitions must at least describe what "kind" it is and the "config" for that kind.'

        kind = definition['kind']
        config = getattr(self, '_configure_default_%s' % kind)(variable)
        config.update(definition.get('config', {}))

        function = getattr(self, '_configure_function_%s' % kind)(**config)
        
        return self._close_function(function, definition.get('alias', {}))
    
    
    def _close_function(self, function, aliases=None):
        """
        Create closures for each variable's function.
        This will ensure each step is resolved correctly.
        """
        return WrappedSimulationFunction(self, function, aliases)
    

    def _initialize_transitions(self, transition_definitions):        
        transitions = []
        
        for source, destinations in transition_definitions.items():
            transition = {
                'source': source,
                'trigger': self._TRANSITION_CHECK,
            }
            
            if isinstance(destinations, (str,unicode)):
                transition['dest'] = destinations
            else:
                for dest, conditionals in destinations.items():
                    transition['dest'] = dest
                    
                    for check, conditional in conditionals.items():
                        if isinstance(conditional, (list,tuple,set)):
                            transition[check] = [
                                self._initialize_conditional(condition)
                                for condition 
                                in conditional]
                        else:
                            transition[check] = self._initialize_conditional(conditional)                       
            
            transitions.append(transition)
        
        self._end_states = [state for state 
                            in self._state_variable_definitions
                            if not state in transition_definitions]
        
        if self._end_states:
            transitions.append({
                'source': self._end_states,
                'dest': '=',
                'trigger': 'check_state'
            })
        
        self._transition_definitions = transitions
    
    
    def _initialize_conditional(self, conditional):
        """Conditions trigger on match"""
        if isinstance(conditional, dict):
            def check_values(self=self, values=conditional.copy()):
                return all(
                        self._variables[variable] == value
                        for variable, value
                        in values.items()
                    )
            return check_values
            
        raise NotImplementedError("No conditional resolver found for '%r'" % conditional)

    
    def _reconfigure_after_state_change(self, *args, **kwargs):
        self._variables['_n_states'] += 1
        self._variables['_n_steps'] = 0
        if self._escapement_variable:
            self._variables['_t_state'] = self._variables[self._escapement_variable]
        self._definitions = self._state_variable_definitions[self.state]
        self._initialize_variables()
        
        
    def step(self):
        """
        Incrememnt the simulation variables by one iteration.
        Note: this always acts on the last step's values.
        """
        self._variables['_n_steps'] += 1
        
        if self._escapement:
            t_prev = self._variables[self._escapement_variable]
            self._escapement()
            self._variables['_t_step'] = self._variables[self._escapement_variable] - t_prev
        
        new_values = {}
        
        for variable, function in self._functions.items():
            new_values[variable] = function()
            
        self._variables.update(new_values)
        
        self.check_state()
        
        
    def __repr__(self):
        max_state_len = max(len(s) for s in self.states.keys())
        format_string = '<Simulation: [%%%ds] {%%s}>' % max_state_len
        return format_string % (self.state, ', '.join(['%s: %s' % (k,v) for k,v 
                                                       in sorted(self._variables.items())
                                                       if not k.startswith('_')
                                                      ]))


def load_simulator(definition, mixins_package='ligature.simulators.mixins'):
        
    configuration = yaml_loader(definition, FullLoader)
    configuration['raw_definition'] = definition
    
    mixins = []
    mixins += [
            getattr(__import__('%s.%s' % (mixins_package, mixin.lower()), 
                               fromlist=['%sMixin' % mixin]), 
                    '%sMixin' % mixin)
            for mixin in configuration.pop('mixins')
        ]
    
    mixins += [Process]
    
    return type('Simulator', tuple(mixins), {})(**configuration)
    



import unittest


from ligature.recordset import RecordSet

from ligature.transforms.merge import Merge


class MergeTestCase(unittest.TestCase):

	def test_basic(self):

		rsa = [(1,2,3,4),(5,6),(7,8,9)]
		rsb = [(0,1,0,1),(0,1),(0,1,0)]
		rsc = [(9,8,7,6,5),(4,3,2),(1,)]
		rsd = [(1,0,1,0,1),(0,1,0),(1,)]

		rs1 = RecordSet(recordType='ab')
		for g in zip(rsa,rsb):
		    rs1.append(v for v in zip(*g))
		
		rs2 = RecordSet(recordType='cb')
		for g in zip(rsc,rsd):
		    rs2.append(v for v in zip(*g))

		merge = Merge([rs1, rs2])

		self.assertEqual(
			merge.results._RecordType._fields,
			('a', 'b', 'c')
			)

		self.assertEqual(
			merge.results._groups[0][0]._tuple,
			(1, 1, 9) 
			)

		self.assertEqual(
			[[record._tuple for record in group] for group in merge],
			[[(1, 1, 9),
			  (2, 0, 8),
			  (3, 1, 7),
			  (4, 0, 6),
			  (5, 1, 5),
			  (6, 0, 4),
			  (7, 1, 3),
			  (8, 0, 2),
			  (9, 1, 1)]]
			)



def runTests():
	suite = unittest.TestLoader().loadTestsFromTestCase(MergeTestCase)
	unittest.TextTestRunner(verbosity=2).run(suite)



if __name__ == '__main__':
    unittest.main()


from .element import ElementScanner


class RecordScanner(ElementScanner):
    """Returns the whole record when emitting."""
    def __iter__(self):
        self._pending_finally()
        for group in self.source._groups[self._group_cursor:]: # error here merely stops iteration
            self._group_cursor += 1
            for record in self._iterGroup(group):
                yield record


import unittest
import math

from ligature.expression import convert_to_postfix, Expression


class ExpressionTestCase(unittest.TestCase):


	def test_parsing(self):
		pass


	def test_tupleing(self):

		self.assertEqual(Expression( '1,2' )(), (1,2) )
		self.assertEqual(Expression( '1,2,3,4,5' )(), (1,2,3,4,5) )
		self.assertEqual(Expression('1,2,(3,4),5')(), (1, 2, (3, 4), 5) )


	def test_passthru(self):

		self.assertEqual(Expression( 'math.pi' )(), math.pi)
		self.assertEqual(Expression( '22.37' )(), 22.37)
		self.assertEqual(Expression( 'a' )(12), 12)
		self.assertEqual(Expression( 'a' )(a=12), 12)


	def test_precedence(self):

		# Ensure commutive properties are properly parsed
		self.assertEqual(Expression( 'c + a - b' )(a=1,b=2,c=4), 3)
		self.assertEqual(Expression( 'c - b + a' )(a=1,b=2,c=4), 3)

		# Ensure algebraic precedence is held
		self.assertEqual(Expression( 'a + b * c' )(a=1,b=2,c=4), 9)
		self.assertEqual(Expression( 'b * c + a' )(a=1,b=2,c=4), 9)

	def test_parentheticals(self):

		testStack = ((2, '1'), (2, '2'), (2, '3'), (51, '*'), (51, '+'))

		expressions = [ '1 + 2 * 3', 
						'1 + (2 * 3)',
						'(1 + (2 * 3))',
						'(1 + ((2) * 3))'
						]
		for expression in expressions:
			self.assertEqual(testStack, convert_to_postfix(expression))

	def test_externals(self):
		
		self.assertEqual(Expression('max(a)')([1,2,3,8,3,2,-1]), 8)
		self.assertEqual(Expression('min(a)')([1,2,3,8,3,2,-1]), -1)

		self.assertEqual(Expression('math.cos(math.pi)')(), -1.0)


	def test_arguments(self):

		expression = Expression( 'a + b * c' )

		self.assertEqual(expression(1,2,4), 9)

		# Specific overrides
		self.assertEqual(expression(1,2,c=4), 9)
		self.assertEqual(expression(1,2,b=3,c=4), 13)
		with self.assertRaises(IndexError):
			_ = expression(1,2,b=3) # missing c!
		self.assertEqual(expression(1,2,4,b=3), 13)
		self.assertEqual(expression(a=1,b=2,c=4), 9)

		# expansion semantics
		self.assertEqual(expression(*(1,2,4)), 9)
		self.assertEqual(expression(**{'a':1,'b':2,'c':4}), 9)
		self.assertEqual(expression(**{'a':1,'b':2,'c':4}), 9)
		self.assertEqual(expression(**dict(a=1,b=2,c=4)), 9)
		self.assertEqual(expression(*(1,2,4),**dict(a=10,b=20,c=40)), 810)


	def test_safety(self):
		pass



def runTests():
	suite = unittest.TestLoader().loadTestsFromTestCase(ExpressionTestCase)
	unittest.TextTestRunner(verbosity=2).run(suite)



if __name__ == '__main__':
    unittest.main()


from ..transform import Transform
from ..recordset import RecordSet
from ..scanners.replaying import ReplayingGroupScanner, ReplayingRecordScanner

from itertools import izip as zip


class Regroup(Transform):
    """Take the first recordset's grouping and enforce it on another's.
    NOTE: This assumes this can work! No validation (yet) is done.
      Rather, this is a way to flag that two recordsets are aligned.
    """
    ScanClass = (ReplayingGroupScanner, ReplayingRecordScanner)
    
    def __init__(self, source, target):
        # Initialize mixins
        super(Regroup, self).__init__()
        
        self.sources = (source, target)
        self._resultSet = RecordSet(recordType=target._RecordType)
        self._generateScanners()
    
    def _generateScanners(self):
        #source, target = self.sources
        #self.scanners = (GroupScanner(source), RecordScanner(target))
        self.scanners = tuple(sc(s) for sc,s in zip(self.ScanClass, self.sources))
        
    def transform(self):
        source, target = self.scanners
        
        for group in source:
            newGroup = tuple(record for _,record in zip(group, target))
            if len(newGroup) == len(group):
                self._resultSet.extend( (newGroup,) )
                source.anchor()
                target.anchor()
            else:
                break



from weakref import WeakSet


class UpdateModel(object):
    """Provide the object with a way to notify other objects
      that depend on it.
    """
    
    # Slots ensures we're explicit and fast
    __slots__ = ('_sources', '_listeners')
    
    def __init__(self, *args, **kwargs):
        """Initialize the chain.
        By tracking both sources and listeners, we can make a graph of what
          gets updated by what.
        """
        # Initialize mixins - NOPE Update is not cooperative.
        # It expects to be a base class
        #super(UpdateModel, self).__init__(*args, **kwargs)
        self._sources = tuple()
        self._listeners = WeakSet()
        
    def subscribe(self, listener):
        """Add a listener to the subscriber list.
        This isn't a set - order will likely help efficiency,
          the list will be updated infrequently, and the list
          should never get very big anyhow.
        Note that Calc objects have a source to act as their publisher list.
          (In case we want to backtrace.)
        """
        if not listener in self._listeners:
            self._listeners.add(listener)
    
    def unsubscribe(self, listener):
        """Remove a listener from the subscriber list.
        """
        while listener in self._listeners:
            self._listeners.remove(listener)
    
    def notify(self, oldSelector, newSelector):
        """Fires an update to make sure dependents are updated, if needed.
        The selectors show what happened in the update.
        """
        for dependent in self._listeners:
            try:
                dependent.update(oldSelector, newSelector)
            except NotImplementedError:
                pass
            except AttributeError:
                pass
            
    def update(self, oldSelector, newSelector):
        """Execute the update. Each class will have its own way to implement this."""
        # (None, None) signals that the data is out of date, 
        #  but there is nothing for dependents to do yet.
        #self._needsUpdate = True
        # Pass-through updates without triggering
        self.notify(None,None)
    
    @property
    def listeners(self):
        return self._listeners
    
    @listeners.setter
    def listeners(self, newListeners):
        self._replace_listeners(newListeners)
    
    def _replace_listeners(self, newSources):        
        """If the listeners are changed en masse, break
           all the subscriptions.
           This setter makes sure the subscription methods are never skipped.
        """
        while self._listeners:
            listener = self._listeners[0]
            self.unsubscribe(listener)
            
        for listener in newListeners:
            self.subscribe(listener)
        
    @property
    def sources(self):
        return self._sources
    
    @sources.setter
    def sources(self, newSources):
        self._replace_sources(newSources)
    
    def _replace_sources(self, newSources):
        for source in set(self._sources).difference(set(newSources)):
            source.unsubscribe(self)
        for source in newSources:
            source.subscribe(self)
        self._sources = newSources



from ..transform import Transform
from ..recordset import RecordSet
from ..scanners.record import RecordScanner

from itertools import izip as zip


class LagBucket(Transform):
    """Creates a new recordset of tuples of each of the given records,
    with the first being from `lag` records back.
    """
    __slots__ = ('_lag', '_lagRecords')
    ScanClass = RecordScanner
    
    def __init__(self, source, lag=1):
        #Initialize mixins
        super(LagBucket, self).__init__()
        
        self._lag = lag
        self.sources = (source,)
        self._resultSet = RecordSet(recordType=source._RecordType)
        self.scanners = (self.ScanClass(source),)
        self._lagRecords = []
        
    def transform(self):
        while len(self._lagRecords) < self._lag:
            self._lagRecords.append(next(self.scanners[0]))
        else:
            for record in self.scanners[0]:
                prev = self._lagRecords.pop(0)
                self._resultSet.append(
                    # cast to the record early so the tuples are not misunderstood
                        tuple(
                            (last,this)
                            for last,this
                            in zip(prev._tuple,record._tuple) ) )
                self._lagRecords.append(record)


"""A simple enumeration class.
"""


__copyright__ = """Copyright (C) 2020 Corso Systems"""
__license__ = 'Apache 2.0'
__maintainer__ = 'Andrew Geiger'
__email__ = 'andrew.geiger@corsosystems.com'

__all__ = ['Enum']


class MetaEnumValue(type):
    
    def __init__(cls, clsname, bases, attributes):
        for base in bases:
            if base.__name__ != 'EnumValue':
                setattr(cls, '_type', base)
                break
        
        def __setattr__readonly__(self, key, _):
            raise AttributeError("<%r> is readonly" % self)
        setattr(cls, '__setattr__', __setattr__readonly__)

        
class EnumValue(object):
    _parent = None
    _type = None
        
    # def __str__(self):
    #     return '%s.%s' % (self._parent.__name__, self.__class__.__name__)

    def __repr__(self):
        return '<%s.%s %s>' % (self._parent.__name__, 
                               self.__class__.__name__, 
                               self._type.__str__(self))
    

    
class MetaEnum(type):
    
    _initFields = ('_fields', '_values')
    _class_initialized = False
        
    def __init__(cls, clsname, bases, attributes):
        
        super(MetaEnum, cls).__setattr__('_class_initialized', False) # bypass interlock
        
        fvs = [(key,value) for key, value in attributes.items() if not key.startswith('_')]
        
        if fvs:
            fields,values = zip(*sorted(fvs, key=lambda (key,value): value))
        
            setattr(cls, '_fields', fields)
            setattr(cls, '_values', values)
            
            for key,value in fvs:
                EnumAttribute = MetaEnumValue(key, (EnumValue,type(value)), {'_parent': cls})
                setattr(cls, key, EnumAttribute(value))

            
        else:
            setattr(cls, '_fields', tuple())
            setattr(cls, '_values', tuple())
        
        cls._class_initialized = True
                
            
    def __setattr__(cls, key, value):
        if cls._class_initialized:
            raise AttributeError("<%s> attributes are readonly" % cls.__name__)
        else:
            super(MetaEnum, cls).__setattr__(key, value)                
                
    def __contains__(cls, enum_key):
        return enum_key in cls._fields
    
    def keys(cls):
        return cls._fields

    def values(cls):
        return cls._values

    def __iter__(cls):
        return iter(getattr(cls, field) for field in cls._fields)
    
    def __getitem__(cls, attribute):
        return getattr(cls, attribute)
    
    def __str__(cls):
        return cls.__name__
    
    def __repr__(cls):
        return "<%s {%s}>" % (cls.__name__, 
                              ', '.join("%s: %s" % (repr(field), repr(value)) 
                                        for field, value in zip(cls._fields, cls._values)))

    
class Enum(object):
    __metaclass__ = MetaEnum
    __slots__ = tuple()
        
    _fields = tuple()
    _values = tuple()
    
    def __new__(cls, value=None):
        if value is not None and value in cls._values:
            return getattr(cls, cls._fields[[i for i,v in enumerate(cls._values) if v == value][0]])
        raise KeyError("%s is an enumeration and %r is not a valid value." % (cls.__name__, value)) 
    
    def __init__(cls):
        raise NotImplementedError("%s is an enumeration and does not support instantiation." % cls.__name__)
        
    def __setattr__(cls, key, value):
        raise AttributeError("<%s> attributes are readonly" % cls.__name__)



import unittest


from ligature.recordset import RecordSet
from ligature.examples import simpleRecordSet, simpleAddition

from ligature.scanners.chunk import ChunkScanner


class ChunkScannerTestCase(unittest.TestCase):

	def test_basic(self):

		srs = RecordSet(simpleRecordSet)
		
		scanner = ChunkScanner(srs,'a')

		# Scanners are like generators...
		self.assertEqual(
			[v for v in scanner], 
			[(1, 2, 3, 4), (5, 6), (7, 8, 9)]
			)

		# ... and will exhaust when fully consumed
		self.assertEqual(
			[v for v in scanner],
			[]
			)

		srs.extend(simpleAddition)

		# adding data means the scanner consumes the new data
		self.assertEqual(
			[v for v in scanner],
			[(11, 12, 13), (14, 15, 16)]
			)

		# resetting the scanner means it will replay the whole dataset
		scanner.reset()

		self.assertEqual(
			[v for v in scanner],
			[(1, 2, 3, 4), (5, 6), (7, 8, 9), (11, 12, 13), (14, 15, 16)]
			)



def runTests():
	suite = unittest.TestLoader().loadTestsFromTestCase(ChunkScannerTestCase)
	unittest.TextTestRunner(verbosity=2).run(suite)



if __name__ == '__main__':
    unittest.main()


import unittest

from ligature.record import genRecordType


class GenerateRecordTypeTestCase(unittest.TestCase):

	def test_instantiation(self):

		# string as a list
		R = genRecordType('abc')
		self.assertEqual(R._fields, ('a', 'b', 'c'))

		# list of columns
		R = genRecordType(['a','b','c'])
		self.assertEqual(R._fields, ('a', 'b', 'c'))

		# from generator
		R = genRecordType((h for h in ['a','b','c']))
		self.assertEqual(R._fields, ('a', 'b', 'c'))

		# from DataSet
		try:
			from ligature.examples import simpleDataset
			R = genRecordType(simpleDataset)
			self.assertEqual(R._fields, ('a', 'b', 'c'))
		except ImportError:
			pass


		# One columned
		R = genRecordType('a')
		r = R(1)
		self.assertEqual(r._tuple, (1,))


	def test_sanitizingFields(self):

		fields = ['Column 1', '3', 'Col@2!']
		R =  genRecordType(fields)
		self.assertEqual(
			R._lookup, 
			{'3': 1, 'C3': 1, 'Col@2!': 2, 'Col_2_': 2, 'Column 1': 0, 'Column_1': 0} )


	def test_replacement(self):

		R = genRecordType('abc')
		r = R((1,2,3))
		r._replace(b=5)

		self.assertEqual(r._asdict(), {'a': 1, 'b': 5, 'c': 3} )


	def test_interfaces(self):

		# create the example 
		R = genRecordType('abc')
		r = R( (1,2,3) )

		self.assertEqual(r._asdict(), {'a': 1, 'b': 2, 'c': 3} )
		self.assertEqual(r._tuple, (1, 2, 3))
		self.assertEqual(r.values, (1, 2, 3))


def runTests():
	suite = unittest.TestLoader().loadTestsFromTestCase(GenerateRecordTypeTestCase)
	unittest.TextTestRunner(verbosity=2).run(suite)



if __name__ == '__main__':
    unittest.main()


import unittest


from ligature.recordset import RecordSet
from ligature.examples import simpleRecordSet, simpleAddition

from ligature.scanners.group import GroupScanner


class GroupScannerTestCase(unittest.TestCase):

	def test_basic(self):

		srs = RecordSet(simpleRecordSet)
		
		scanner = GroupScanner(srs,'a')

		# Scanners are like generators...
		self.assertEqual(
			[[record._tuple for record in group] for group in scanner], 
			[[(1, 0), (2, 1), (3, 0), (4, 1)], [(5, 0), (6, 1)], [(7, 0), (8, 1), (9, 0)]]
			)

		# ... and will exhaust when fully consumed
		self.assertEqual(
			[group for group in scanner],
			[]
			)

		srs.extend(simpleAddition)

		# adding data means the scanner consumes the new data
		self.assertEqual(
			[[record._tuple for record in group] for group in scanner],
			[[(11, 1), (12, 0), (13, 1)], [(14, 0), (15, 1), (16, 0)]]
			)

		# resetting the scanner means it will replay the whole dataset
		scanner.reset()

		self.assertEqual(
			[len(group) for group in scanner], 
			[4, 2, 3, 3, 3]
			)



def runTests():
	suite = unittest.TestLoader().loadTestsFromTestCase(GroupScannerTestCase)
	unittest.TextTestRunner(verbosity=2).run(suite)



if __name__ == '__main__':
    unittest.main()


import unittest


from ligature.recordset import RecordSet
from ligature.examples import simpleRecordSet, simpleAddition

from ligature.transforms.pivot import Pivot


class PivotTestCase(unittest.TestCase):

	def test_basic(self):

		srs = RecordSet(simpleRecordSet)
		
		pivot = Pivot(srs)

		# Pivot converts a group or records into one record per group
		self.assertEqual(
			[[record._tuple for record in group] for group in pivot], 
			[[((1, 2, 3, 4), (0, 1, 0, 1))], 
			 [((5, 6), (0, 1))], 
			 [((7, 8, 9), (0, 1, 0))]]
			)

		srs.extend(simpleAddition)

		# adding data means the transform consumes the new data when checked
		self.assertEqual(
			[[record._tuple for record in group] for group in pivot],
			[[((1, 2, 3, 4), (0, 1, 0, 1))],
			 [((5, 6), (0, 1))],
			 [((7, 8, 9), (0, 1, 0))],
			 [((11, 12, 13), (1, 0, 1))],
			 [((14, 15, 16), (0, 1, 0))]]
			)



def runTests():
	suite = unittest.TestLoader().loadTestsFromTestCase(PivotTestCase)
	unittest.TextTestRunner(verbosity=2).run(suite)



if __name__ == '__main__':
    unittest.main()


"""
	Mixin function template and support for Process simulation
"""


class MetaFunctionMixin(type):
	
	_meta_config_methods = ('_configure_default_',
							'_configure_function_',
						   )
		
	def __new__(cls, clsname, bases, attributes):
		
		if any((
			clsname == 'MixinFunctionSupport',
			not clsname.endswith('Mixin'),
		)):
			return super(MetaFunctionMixin, cls).__new__(cls, clsname, bases, attributes)
			
		support_name = clsname[:-len('Mixin')] #  clsname.partition('_')[2]
		
		for method_name in cls._meta_config_methods:
			
			specialized_method_name = '%s%s' % (method_name, support_name)
			if method_name in attributes:
				method = attributes.pop(method_name)
			else:
				raise NotImplementedError('MixinFunctionSupport subclasses must at least cover the default config methods: %r' % cls._meta_config_methods)
			attributes[specialized_method_name] = method
		
		return super(MetaFunctionMixin, cls).__new__(cls, clsname, bases, attributes)


class MixinFunctionSupport(object):
	__metaclass__ = MetaFunctionMixin
		
	def _configure_default_(self, variable):
		# by default pass along enough info to come up with something useful
		return dict(variable=variable, value=self._DEFAULT_START_VALUE)

	def _configure_function_(self, **configuration):
		if 'hold' in configuration:
			def hold_value(self=self, variable=configuration['variable']):
				return self._variables[variable]
			return hold_value
		elif 'value' in configuration:
			def static_value(self=self, value=configuration['value']):
				return value
			return static_value
		elif 'function' in configuration:
			return configuration['function']
		raise NotImplementedError('Mixins should explicitly configure functions.')
		
	def _resolve_arguments(self, some_callable):
		return super(MixinFunctionSupport, self)._resolve_arguments(some_callable)

	def _resolve_variable_definition(self, variable_definition):
		return super(MixinFunctionSupport, self)._resolve_variable_definition(variable_definition)



from ..scanners.element import ElementScanner
from ..transform import Transform
from ..recordset import RecordSet
from ..compose import Composable
from ..record import genRecordType

from itertools import izip as zip


class Merge(Transform):
    """Combine the source recordsets into one recordset.
    The new record type will have all the source columns,
      with the caveat that later sources win for overlaps.
    """
    ScanClass = ElementScanner
    
    def __init__(self, sources):
        # Initialize mixins
        super(Merge, self).__init__()
        self.sources = tuple(sources)
        self._resolveSources()
        
    
    def _resolveSources(self):
        """Sources may overlap: if so, only take the latter."""        
        rawSources = [source.results if isinstance(source, Composable) else source
                      for source 
                      in self.sources]
        
        allFields = []
        # Gather all the fields
        for source in rawSources:
            for field in source._RecordType._fields:
                allFields.append(field)
        
        scanners = []
        sourceFields = set(allFields)
        for source in reversed(rawSources):
            for field in source._RecordType._fields:
                if field in sourceFields:
                    sourceFields.remove(field)
                    scanners.append( (field, self.ScanClass(source, field)) )
                if not sourceFields:
                    break
            if not sourceFields:
                break
            
        # While we want to prioritize later sources, the fields should
        #   likely keep the same order, starting with the earlier sources.
        # see https://stackoverflow.com/a/12814719/1943640
        scanners.sort(key=lambda entry: allFields.index(entry[0]))
        
        self.scanners = tuple(scanner 
                              for field,scanner 
                              in scanners)
        self._resultSet = RecordSet(recordType=genRecordType(field 
                                                             for field,scanner 
                                                             in scanners))
        
    def transform(self):
        """Simply scan down the sources, generating new records."""
        self._resultSet.append( tuple(
            self._resultSet.coerceRecordType(newRecordValues)
            for newRecordValues
            in zip(*self.scanners)
            ) )        



from ..calculation import Calculation
from ..scanners.chunk import ChunkScanner
from ..calculations.sweep import Sweep


class Window(Sweep, Calculation):

    ScanClass = ChunkScanner
    
    def calculate(self):
        """Run the aggregate function by group creating one new group.
           If groups don't matter after windowing, this is easiest.
           Each resulting group is an update.
        f(a,b)=sum(a)-sum(b) 
        rs.a = [(1,2,3,4),(5,6),(7,8,9)]
        rs.b = [(0,1,0,1),(0,1),(0,1,0)]
        calc = [(8,10,23)]               # 1 group of 3
        """
        super(Window, self).calculate()


import unittest


from ligature.recordset import RecordSet
from ligature.examples import simpleRecordSet, simpleAddition

from ligature.scanners.element import ElementScanner


class ElementScannerTestCase(unittest.TestCase):

	def test_basic(self):

		srs = RecordSet(simpleRecordSet)
		
		scanner = ElementScanner(srs,'a')

		# Scanners are like generators...
		self.assertEqual(
			[v for v in scanner], 
			[1, 2, 3, 4, 5, 6, 7, 8, 9]
			)

		# ... and will exhaust when fully consumed
		self.assertEqual(
			[v for v in scanner],
			[]
			)

		srs.extend(simpleAddition)

		# adding data means the scanner consumes the new data
		self.assertEqual(
			[v for v in scanner],
			[11, 12, 13, 14, 15, 16]
			)

		# resetting the scanner means it will replay the whole dataset
		scanner.reset()

		self.assertEqual(
			[v for v in scanner],
			[1, 2, 3, 4, 5, 6, 7, 8, 9, 11, 12, 13, 14, 15, 16]
			)



def runTests():
	suite = unittest.TestLoader().loadTestsFromTestCase(ElementScannerTestCase)
	unittest.TextTestRunner(verbosity=2).run(suite)



if __name__ == '__main__':
    unittest.main()


"""
    Python port of the easings from the excellent resource https://github.com/ai/easings.net
    URL: https://easings.net

    Some minor additions and wrappers made for convenience.    
"""


import math
from itertools import permutations
from .enum import Enum
  
class DIRECTION(Enum):
    IN = 'in'
    OUT = 'out'
    IN_OUT = 'in_out'

class ALGORITHM(Enum):
    
    QUADRATIC = 'quad' # Quadratic - squared
    CUBIC = 'cubic'    # Cubic - x ^ third
    QUARTIC = 'quart'  # Quartic - x to the four
    QUINTIC = 'quint'  # Quintic - x's fifth power

    LINEAR = 'linear'
    SINE = 'sine'
    EXPONENT = 'expo'
    CIRCULAR = 'circ'
    BACK = 'back'
    ELASTIC = 'elastic'
    
    
__license__ = 'GPLv3'

# Modifications and additions by:
__maintainer__ = 'Andrew Geiger'
__email__ = 'andrew.geiger@corsosystems.com'


class MetaEaseFunctions(type):
    
    def __init__(cls, clsname, bases, attributes):
        
        for power, algo in enumerate(cls._algebraic):
            for direction in cls._directions:
                powerFunction = getattr(cls, '%s_power' % direction)
                def powerClosure(cls, x, p=power+2, pfun=powerFunction):
                    return pfun(x,p)
                setattr(cls, '%s_%s' % (direction, algo), classmethod(powerClosure))
        
        for firstList,secondList in permutations([cls._directions,cls._algos]):
            for first in firstList:
        
                class EaseChain(object):
                    pass

                for second in secondList:
                    try:
                        function = getattr(cls, '%s_%s' % (first, second))
                    except AttributeError:
                        function = getattr(cls, '%s_%s' % (second,first))

                    @classmethod
                    def closure(_, x, fun=function):
                        return fun(x)

                    setattr(EaseChain, second.upper(), closure)

                setattr(cls, first.upper(), EaseChain)
               
        return super(MetaEaseFunctions, cls).__init__(clsname, bases, attributes)
    

    
class EaseFunctions(object):
    """An easing object.
    Transliterated from 
    https://github.com/ai/easings.net/blob/master/src/easings/easingsFunctions.ts
    """
    __metaclass__ = MetaEaseFunctions

    _algebraic = ('quad', 'cubic', 'quart', 'quint')
    _algos = _algebraic + ('linear', 'sine', 'expo', 'circ', 'back', 'elastic')
    _directions = ('in', 'out', 'in_out')

    
    _c1 = 1.70158
    _c2 = _c1 * 1.525
    _c3 = _c1 + 1
    _c4 = (2 * math.pi) / 3
    _c5 = (2 * math.pi) / 4.5


    @classmethod
    def in_linear(cls, x):
        return x
    
    @classmethod
    def out_linear(cls, x):
        return x

    @classmethod
    def in_out_linear(cls, x):
        return x

    
    @classmethod
    def in_power(cls, x, power):
        return math.pow(x, power)
    
    @classmethod
    def out_power(cls, x, power):
        return 1 - math.pow(1-x, power)
    
    @classmethod
    def in_out_power(cls, x, power):
        if x < 0.5:
            return 2*(power-1)*math.pow(x,power)
        else:
            return 1 - math.pow(-2*x + 2, power) / 2.0
    
    
    @classmethod
    def in_sine(cls, x):
        return 1 - math.cos((x*math.pi)/2.0)
    
    @classmethod
    def out_sine(cls, x):
        return math.sin((x*math.pi)/2.0)
    
    @classmethod
    def in_out_sine(cls, x):
        return -(math.cos(x*math.pi) - 1)/2.0
    
    
    @classmethod
    def in_expo(cls, x):
        if x == 0:
            return 0
        return math.pow(2, 10*x - 10)
        
    @classmethod
    def out_expo(cls, x):
        if x == 1:
            return 1
        return 1 - math.pow(2, -10*x)
        
    @classmethod
    def in_out_expo(cls, x):
        if x == 0:
            return 0
        if x == 1:
            return 1
        if x < 0.5:
            return math.pow(2, 20*x - 10)/2.0
        else:
            return (2 - math.pow(2, -20*x + 10))/2.0
    
    
    @classmethod
    def in_circ(cls, x):
        return 1 - math.sqrt(1 - math.pow(x, 2))
    
    @classmethod
    def out_circ(cls, x):
        return math.sqrt(1 - math.pow(x - 1, 2))
    
    @classmethod
    def in_out_circ(cls, x):
        if x < 0.5:
            return (1 - math.sqrt(1 - math.pow(2*x, 2)))/2.0
        else:
            return (math.sqrt(1 - math.pow(-2*x + 2, 2)) + 1)/2.0
        
        
    @classmethod
    def in_back(cls, x):
        return cls._c3*math.pow(x, 3) - cls._c1*math.pow(x, 2)
    
    @classmethod
    def out_back(cls, x):
        return 1 + cls._c3*math.pow(x-1, 3) + cls._c1*math.pow(x-1, 2)
    
    @classmethod
    def in_out_back(cls, x):
        if x < 0.5:
            return (math.pow(2*x, 2)*((cls._c2 + 1)*2*x - cls._c2))/2.0
        else:
            return (math.pow(2*x - 2, 2)*((cls._c2 + 1)*(x*2 - 2) + cls._c2) + 2)/2.0
        
    
    @classmethod
    def in_elastic(cls, x):
        if x == 0:
            return 0
        if x == 1:
            return 1
        return -math.pow(2, 10*x - 10) * math.sin((x*10 - 10.75) * cls._c4)
    
    @classmethod
    def out_elastic(cls, x):
        if x == 0:
            return 0
        if x == 1:
            return 1
        return math.pow(2, -10*x) * math.sin((x*10 - 0.75) * cls._c4) + 1

        
    @classmethod
    def in_out_elastic(cls, x):
        if x == 0:
            return 0
        if x == 1:
            return 1
        if x < 0.5:
            return -(math.pow(2, 20*x - 10) * math.sin((20*x - 11.125) * cls._c5))/2.0
        else:
            return (math.pow(2, -20*x + 10) * math.sin((20*x - 11.125) * cls._c5))/2.0 + 1
        
    
    @classmethod
    def in_bounce(cls, x):
        return 1 - cls.out_bounce(1-x)
    
    @classmethod
    def out_bounce(cls, x):
        n = 7.5625
        d = 2.75
        
        if x < (1.0/d):
            c = 0
            
        elif x < (2.0/d):
            x -= 1.5/d
            c = 0.75
            
        elif x < (2.5/d):
            x -= 2.25/d
            c = 0.9375
            
        else:
            x -= 2.625/d
            c = 0.984375
            
        return n1 * x * x + c

    @classmethod
    def in_out_bounce(cls, x):
        if x < 0.5:
            return (1 - cls.out_bounce(1 - 2*x))/2.0
        else:
            return (1 + cls.out_bounce(2*x - 1))/2.0
        
        
class Easing(object):
    __slots__ = ('function', 
                 'start', 'finish', 
                 'time_start', 'time_end', 'steps')
    
    def __init__(self, 
                 ease_type=ALGORITHM.LINEAR,
                 direction=DIRECTION.IN,
                 start=0.0,
                 finish=1.0,
                 steps=None,
                 time_start=0.0, 
                 time_end=None,
                 duration=None,
                 ):
        """
        Ease from start to finish.
        
        If steps are an integer, it's assumed to iterate that many times.
        If steps are a float, it's assumed the steps run until time_end.
        If the steps are undefined, it's assumed the param is in relation
          to time_start and time_end.
          
        Remember the fencepost problem: steps are the fence, not the posts. 
          You start at the beginning - the first step is the first increment.
          Thus for steps=10, the iterable will yield 11 times!
        """
        
        self.function = getattr(EaseFunctions, '%s_%s' % (direction, ease_type))
        
        self.start = start * 1.0
        self.finish = finish * 1.0

        if duration and not time_end:
            time_end = time_start + duration
        
        assert time_start < time_end, "Time must flow forward (though scale may not)"
        self.time_start = time_start * 1.0
        self.time_end = time_end * 1.0
        self.steps = steps or None
        
    @property
    def step_by_count(self):
        return self.steps and isinstance(self.steps, (int, long))
    @property
    def step_by_increment(self):
        return self.steps and isinstance(self.steps, (float,))
        
    @property
    def span(self):
        return self.finish - self.start
    
    @property
    def time_span(self):
        return self.time_end - self.time_start

    @property
    def time_bounds(self):
        return slice(self.time_start, self.time_end, self.steps)
    
    
    def normalize_time(self, t):
        if self.steps:
            if self.step_by_count:
                if t <= 0:
                    return 0.0
                elif t >= self.steps:
                    return 1.0
                else:
                    return (t * 1.0) / self.steps
            elif self.step_by_increment:
                total_inc = (t * 1.0) * self.steps
                if total_inc <= 0:
                    return 0.0
                elif total_inc >= self.time_span:
                    return 1.0
                else:
                    return (total_inc / self.time_span)
        else:
            if t <= self.time_start:
                return 0.0
            elif t >= self.time_end:
                return 1.0
        return (t - self.time_start) / self.time_span

    def interpolate_scale(self, fraction):
        return self.start + (fraction * (self.span))
    
    
    @property
    def scale_bounds(self):
        return slice(self.start, self.finish)
        
    def __call__(self, t):
        t_norm = self.normalize_time(t)
        y_norm = self.function(t_norm)
        return self.interpolate_scale(y_norm)
    
    def __iter__(self):
        assert self.steps, "Cannot iterate without a steps defined."
        if self.step_by_count:
            for i in range(self.steps):
                yield i, self(i)
            else:
                yield self.steps, self.finish
        elif self.step_by_increment:
            for t in range(int(self.time_span/self.steps)):
                yield (t * self.steps) + self.time_start, self(t)
            else:
                if (t * self.steps) + self.time_start < self.time_end:
                    yield self.time_end, self.finish
                    



import unittest


from ligature.recordset import RecordSet
from ligature.examples import simpleRecordSet, simpleAddition

from ligature.scanners.record import RecordScanner


class RecordScannerTestCase(unittest.TestCase):

	def test_basic(self):

		srs = RecordSet(simpleRecordSet)
		
		scanner = RecordScanner(srs,'a')

		# Scanners are like generators...
		self.assertEqual(
			[record._tuple for record in scanner], 
			[(1, 0), (2, 1), (3, 0), (4, 1), (5, 0), (6, 1), (7, 0), (8, 1), (9, 0)]
			)

		# ... and will exhaust when fully consumed
		self.assertEqual(
			[record for record in scanner],
			[]
			)

		srs.extend(simpleAddition)

		# adding data means the scanner consumes the new data
		self.assertEqual(
			[record._tuple for record in scanner],
			[(11, 1), (12, 0), (13, 1), (14, 0), (15, 1), (16, 0)]
			)

		# resetting the scanner means it will replay the whole dataset
		scanner.reset()

		self.assertEqual(
			[record._tuple for record in scanner], 
			[(1, 0), (2, 1), (3, 0), (4, 1), (5, 0), (6, 1), 
			 (7, 0), (8, 1), (9, 0), (11, 1), (12, 0), 
			 (13, 1), (14, 0), (15, 1), (16, 0)]
			)



def runTests():
	suite = unittest.TestLoader().loadTestsFromTestCase(RecordScannerTestCase)
	unittest.TextTestRunner(verbosity=2).run(suite)



if __name__ == '__main__':
    unittest.main()


import unittest


from ligature.recordset import RecordSet

from ligature.transforms.regroup import Regroup


class RegroupTestCase(unittest.TestCase):

	def test_basic(self):

		rsa = [(1,2,3,4),(5,6),(7,8,9)]
		rsb = [(0,1,0,1),(0,1),(0,1,0)]
		rsc = [(1,2,3),(4,),(5,6,7,8,9)]
		rsd = [(0,1,0),(1,),(0,1,0,1,0)]

		# source
		rss = RecordSet(recordType='ab')
		for g in zip(rsa,rsb):
		    rss.append(v for v in zip(*g))

		# target
		rst = RecordSet(recordType='ef')
		for g in zip(rsc,rsd):
		    rst.append(v for v in zip(*g))

		regroup = Regroup(rss, rst)

		# verify it has the same columns as the target
		self.assertEqual(
			regroup.results._RecordType._fields,
			('e', 'f')
			)

		self.assertEqual(
			[[record._tuple for record in group] for group in regroup],
			[[(1, 0), (2, 1), (3, 0), (4, 1)], 
			 [(5, 0), (6, 1)], 
			 [(7, 0), (8, 1), (9, 0)]] 
			)

	def test_misalignment_1(self):

		# source
		rsa = [(1,2,3,4),(5,6),(7,8,9)]
		rsb = [(0,1,0,1),(0,1),(0,1,0)]
		rss = RecordSet(recordType='ab')
		for g in zip(rsa,rsb):
		    rss.append(v for v in zip(*g))

		# target - shorter
		rsa = [(1,2,3),(4,),(5,6,7)]
		rsb = [(0,1,0),(1,),(0,1,0)]
		rst = RecordSet(recordType='ef')
		for g in zip(rsa,rsb):
		    rst.append(v for v in zip(*g))

		regroup = Regroup(rss, rst)


		# verify it has the same columns as the target
		self.assertEqual(
			regroup.results._RecordType._fields,
			('e', 'f')
			)

		# Up to 7 records can be grouped. The final source group
		#   must be incomplete, and is omitted
		self.assertEqual(
			[[record._tuple for record in group] for group in regroup],
			[[(1, 0), (2, 1), (3, 0), (4, 1)], 
			 [(5, 0), (6, 1)]]
			)

		# adding three more means the target has at least enough to complete
		rst.append( [(8,1),(9,0)] )

		# note that the last is omitted, though, 
		#   since the source doesn't have a group to map to it
		self.assertEqual(
			[[record._tuple for record in group] for group in regroup],
			[[(1, 0), (2, 1), (3, 0), (4, 1)], 
			 [(5, 0), (6, 1)], 
			 [(7, 0), (8, 1), (9, 0)]]
			)		


	def test_misalignment_2(self):

		# source
		rsa = [(1,2,3,4)]
		rsb = [(0,1,0,1)]
		rss = RecordSet(recordType='ab')
		for g in zip(rsa,rsb):
		    rss.append(v for v in zip(*g))

		# target - longer
		rsa = [(1,2,3),(4,),(5,6,7),(8,9,10)]
		rsb = [(0,1,0),(1,),(0,1,0),(1,0,10)]
		rst = RecordSet(recordType='ef')
		for g in zip(rsa,rsb):
		    rst.append(v for v in zip(*g))

		regroup = Regroup(rss, rst)


		# verify it has the same columns as the target
		self.assertEqual(
			regroup.results._RecordType._fields,
			('e', 'f')
			)

		# Source only has one group, so that alone gets mapped
		self.assertEqual(
			[[record._tuple for record in group] for group in regroup],
			[[(1, 0), (2, 1), (3, 0), (4, 1)]]
			)

		# adding two more groups...
		rss.extend( [
			 ((5,0),(6,1)),
             ((7,0),(8,1),(9,0))
            ] )

		# ... allows two more groups to be added.
		# Again, note that the last target record is omitted, though, 
		#   since the source doesn't have a group to map to it
		self.assertEqual(
			[[record._tuple for record in group] for group in regroup],
			[[(1, 0), (2, 1), (3, 0), (4, 1)], 
			 [(5, 0), (6, 1)], 
			 [(7, 0), (8, 1), (9, 0)]]
			)		



def runTests():
	suite = unittest.TestLoader().loadTestsFromTestCase(RegroupTestCase)
	unittest.TextTestRunner(verbosity=2).run(suite)



if __name__ == '__main__':
    unittest.main()


import unittest

from ligature.recordset import RecordSet
from ligature.examples import simpleRecordSet, simpleAddition

from ligature.scanners.replaying import ReplayingElementScanner, ReplayingChunkScanner, ReplayingRecordScanner, ReplayingGroupScanner


def sentinel(iterable, stopValue):
	"""A helper to make it simpler to implement sentinel values more idomatically.
	This is a good way to replace a while True loop, removing the need for a break-on-value clause.
	
	>>> [i for i in iter((x for x in range(5)).next, 3)]
	[0, 1, 2]
	>>> [i for i in sentinel(range(5), 3)]
	[0, 1, 2]
	"""
	return iter((x for x in iterable).next, stopValue)



class ReplayingElementScannerTestCase(unittest.TestCase):

	def test_basic(self):

		srs = RecordSet(simpleRecordSet)
		
		scanner = ReplayingElementScanner(srs,'a')

		# Replaying scanners are like generators...
		self.assertEqual(
			[v for v in scanner], 
			[1, 2, 3, 4, 5, 6, 7, 8, 9]
			)

		# ... but will NOT exhaust when fully consumed
		self.assertEqual(
			[v for v in scanner],
			[1, 2, 3, 4, 5, 6, 7, 8, 9]
			)

		# partial iteration...
		self.assertEqual(
			[v for v in sentinel(scanner, 4)],
			[1, 2, 3]
			)
		# ... and an anchor...
		scanner.anchor()
		# ... resumes iteration from the anchor
		self.assertEqual(
			[v for v in scanner],
			[5, 6, 7, 8, 9]
			)

		# Anchoring after iteration stops emitting
		scanner.anchor()
		self.assertEqual(
			[v for v in scanner],
			[]
			)		

		# But if the source adds more data...
		srs.extend(simpleAddition)

		# ... means the scanner consumes the new data
		self.assertEqual(
			[v for v in scanner],
			[11, 12, 13, 14, 15, 16]
			)

		# resetting the scanner means it will replay the whole dataset
		scanner.reset()

		self.assertEqual(
			[v for v in scanner],
			[1, 2, 3, 4, 5, 6, 7, 8, 9, 11, 12, 13, 14, 15, 16]
			)


class ReplayingChunkScannerTestCase(unittest.TestCase):

	def test_basic(self):

		srs = RecordSet(simpleRecordSet)
		
		scanner = ReplayingChunkScanner(srs,'a')

		# Replaying scanners are like generators...
		self.assertEqual(
			[v for v in scanner], 
			[(1, 2, 3, 4), (5, 6), (7, 8, 9)]
			)

		# ... but will NOT exhaust when fully consumed
		self.assertEqual(
			[v for v in scanner],
			[(1, 2, 3, 4), (5, 6), (7, 8, 9)]
			)

		# Anchor after next(iter)
		self.assertEqual(
			next(scanner),
			(1,2,3,4)
			)
		scanner.anchor()

		self.assertEqual(
			[v for v in scanner],
			[(5, 6), (7, 8, 9)]
			)

		scanner.reset()

		# partial iteration with an anchor
		for i,v in enumerate(scanner):
		    if i < 2: 
		        scanner.anchor()
		        break

		self.assertEqual(
			[v for v in scanner],
			[(5, 6), (7, 8, 9)]
			)

		# Anchoring after iteration stops emitting
		scanner.anchor()
		self.assertEqual(
			[v for v in scanner],
			[]
			)		

		# But if the source adds more data...
		srs.extend(simpleAddition)

		# ... means the scanner consumes the new data
		self.assertEqual(
			[v for v in scanner],
			[(11, 12, 13), (14, 15, 16)]
			)

		# resetting the scanner means it will replay the whole dataset
		scanner.reset()

		self.assertEqual(
			[v for v in scanner],
			[(1, 2, 3, 4), (5, 6), (7, 8, 9), (11, 12, 13), (14, 15, 16)]
			)


class ReplayingRecordScannerTestCase(unittest.TestCase):

	def test_basic(self):

		srs = RecordSet(simpleRecordSet)
		
		scanner = ReplayingRecordScanner(srs)

		# Replaying scanners are like generators...
		self.assertEqual(
			[r._tuple for r in scanner], 
			[(1, 0), (2, 1), (3, 0), (4, 1), (5, 0), (6, 1), (7, 0), (8, 1), (9, 0)]
			)

		# ... but will NOT exhaust when fully consumed
		self.assertEqual(
			[r._tuple for r in scanner], 
			[(1, 0), (2, 1), (3, 0), (4, 1), (5, 0), (6, 1), (7, 0), (8, 1), (9, 0)]
			)

		# partial iteration...
		for i,r in enumerate(scanner):
		    if i >= 3:
		        scanner.anchor() # anchor in the iteration
		        break
		# ... and an anchor...
		scanner.anchor()
		# ... resumes iteration from the anchor
		self.assertEqual(
			[r._tuple for r in scanner],
			[(5, 0), (6, 1), (7, 0), (8, 1), (9, 0)]
			)

		# Anchoring after iteration stops emitting
		scanner.anchor()
		self.assertEqual(
			[r._tuple for r in scanner],
			[]
			)		

		# But if the source adds more data...
		srs.extend(simpleAddition)

		# ... means the scanner consumes the new data
		self.assertEqual(
			[r._tuple for r in scanner],
			[(11, 1), (12, 0), (13, 1), (14, 0), (15, 1), (16, 0)]
			)

		# resetting the scanner means it will replay the whole dataset
		scanner.reset()

		self.assertEqual(
			[r._tuple for r in scanner],
			[(1, 0),
			 (2, 1),
			 (3, 0),
			 (4, 1),
			 (5, 0),
			 (6, 1),
			 (7, 0),
			 (8, 1),
			 (9, 0),
			 (11, 1),
			 (12, 0),
			 (13, 1),
			 (14, 0),
			 (15, 1),
			 (16, 0)]
			)


class ReplayingGroupScannerTestCase(unittest.TestCase):

	def test_basic(self):

		srs = RecordSet(simpleRecordSet)
		
		scanner = ReplayingGroupScanner(srs)

		# Replaying scanners are like generators...
		self.assertEqual(
			[[r._tuple for r in g] for g in scanner], 
			[[(1, 0), (2, 1), (3, 0), (4, 1)], 
			 [(5, 0), (6, 1)], 
			 [(7, 0), (8, 1), (9, 0)]]
			)

		# ... but will NOT exhaust when fully consumed
		self.assertEqual(
			[[r._tuple for r in g] for g in scanner], 
			[[(1, 0), (2, 1), (3, 0), (4, 1)], 
			 [(5, 0), (6, 1)], 
			 [(7, 0), (8, 1), (9, 0)]]
			)

		# Anchor after next(iter)
		self.assertEqual(
			[r._tuple for r in next(scanner)],
			[(1, 0), (2, 1), (3, 0), (4, 1)]
			)
		scanner.anchor()

		self.assertEqual(
			[[r._tuple for r in g] for g in scanner],
			[[(5, 0), (6, 1)], 
			 [(7, 0), (8, 1), (9, 0)]]
			)

		scanner.reset()

		# partial iteration with an anchor
		for i,r in enumerate(scanner):
		    if i < 2: 
		        scanner.anchor()
		        break

		self.assertEqual(
			[[r._tuple for r in g] for g in scanner],
			[[(5, 0), (6, 1)], 
			 [(7, 0), (8, 1), (9, 0)]]
			)

		# Anchoring after iteration stops emitting
		scanner.anchor()
		self.assertEqual(
			[v for v in scanner],
			[]
			)		

		# But if the source adds more data...
		srs.extend(simpleAddition)

		# ... means the scanner consumes the new data
		self.assertEqual(
			[[r._tuple for r in g] for g in scanner],
			[[(11, 1), (12, 0), (13, 1)], 
			 [(14, 0), (15, 1), (16, 0)]]
			)

		# resetting the scanner means it will replay the whole dataset
		scanner.reset()

		self.assertEqual(
			[[r._tuple for r in g] for g in scanner],
			[[(1, 0), (2, 1), (3, 0), (4, 1)],
			 [(5, 0), (6, 1)],
			 [(7, 0), (8, 1), (9, 0)],
			 [(11, 1), (12, 0), (13, 1)],
			 [(14, 0), (15, 1), (16, 0)]]
			)



def runTests():
	suite1 = unittest.TestLoader().loadTestsFromTestCase(ReplayingElementScannerTestCase)
	suite2 = unittest.TestLoader().loadTestsFromTestCase(ReplayingChunkScannerTestCase)
	suite3 = unittest.TestLoader().loadTestsFromTestCase(ReplayingRecordScannerTestCase)
	suite4 = unittest.TestLoader().loadTestsFromTestCase(ReplayingGroupScannerTestCase)
	allTests = unittest.TestSuite([suite1, suite2, suite3, suite4])
	unittest.TextTestRunner(verbosity=2).run(allTests)



if __name__ == '__main__':
    unittest.main()


from ..calculation import Calculation
from ..scanners.chunk import ChunkScanner


class Cluster(Calculation):
    
    ScanClass = ChunkScanner
    
    # By group's records
    def calculate(self):
        """For each group, run the function by row, keeping grouping.
           Use this to maintain groupings for later aggregates.
        f(a,b)=a+b 
        rs.a = [(1,2,3,4),(5,6),(7,8,9)]
        rs.b = [(0,1,0,1),(0,1),(0,1,0)]
        calc = [(1,3,3,5),(5,7),(7,9,9)] # 3 groups
        """
        self._resultSet.extend(
            [ tuple( self.function(*rowValues)
                     for rowValues 
                     in zip(*groupedValues) )
             for groupedValues 
             in zip(*self.scanners)
            ])


from .compose import Composable
from .scanner import Scanner

class Transform(Composable):
    """This generally regroups or structures data, but does not 
       calculate new data.
    """    
    ScanClass = Scanner
    
    def _graph_attributes(self):
        label = '%s\\lOut: %s' % (
            type(self).__name__,
            ', '.join(self._resultSet._RecordType._fields))
        return {
            'label': label,
            'shape': 'doubleoctagon'
        }
    
    def _apply(self):
        self.transform()
        self._needsUpdate = False



def passthrough(*args):
    return args


class MetaScanner(type):
    
    registered = {}
    
    def __new__(cls, clsname, bases, attrs):
        newclass = super(MetaScanner, cls).__new__(cls, clsname, bases, attrs)
        cls.registered[clsname] = newclass
        return newclass
    

class Scanner(object):
    """An iterator that holds the position of a cursor while it scans.
    This way, an incompletely scanned recordset can resume once another that's
      zip()'d with it gets more data.
    """
    __metaclass__ = MetaScanner
    __slots__ = ('source', 'getter', 
                 '_group_cursor', '_record_cursor',
                 '_iterating_group', '_iterating_record',
                 )
    
    def __init__(self, source, field=None):
        self.source = source
        
        # allow Scanner to be a bit more generic while useful in the default
        if not field is None:
            self.getter = source._RecordType.getGetter(field)
        else: 
            self.getter = None
            
        self.reset() 
            
    def reset(self):
        # offset low for iteration restarts if more segments get added
        self._group_cursor = 0
        self._record_cursor = 0
        self._iterating_group = None
        self._iterating_record = None

    def _iterGroup(self, group):
        try:
            self._iterating_group = group
            for record in group[self._record_cursor:]:
                # records in groups never get appended, so ensure it never gets over
                self._record_cursor += 1
                yield record
        finally:
            self._iterGroup_finally()

    def _pending_finally(self):
        if self._iterating_group:
            self._iterGroup_finally()
        if self._iterating_record:
            self._iterRecord_finally()

    def _iterGroup_finally(self):
        # In case of incomplete iteration, clean up
        # Note that this is needed, since the finally clause in the iterGroup does NOT run
        #   but once PER ITERATION. That means if more than one scanner is zipped, only
        #   the first finally is run - the rest must know they stopped incompletely!
        if self._record_cursor == len(self._iterating_group):
            self._record_cursor= 0
        else:
            self._group_cursor -= 1
        self._iterating_group = None

    def _iterRecord_finally(self):
        pass    

    def __next__(self):
        return self.next()

    def next(self):
        return next(iter(self))

    @property
    def exhausted(self):
        return self._record_cursor == 0 and self._group_cursor == len(self.source._groups)
    
    def __repr__(self):
        return '%s on group %d and record %d' % (type(self), self._group_cursor, self._record_cursor)
    
    def rewind(self, steps=1):
        raise NotImplementedError("The base scanner class' rewind(steps) must be overridden. This allows accidental consumption to be undone.")
        
    def __iter__(self):
        raise NotImplementedError("The base scanner class' __iter__() must be overridden.")



from .update import UpdateModel

try:
    from networkx import DiGraph
    from nxpd import draw
    
    try:
        if get_ipython().__class__.__name__ == 'ZMQInteractiveShell':
            nxpd.nxpdParams['show'] = 'ipynb'
    except NameError:
        pass
    
    import distutils.spawn, os
    try:
        for binName in ('dot', 'dot.exe'):
            dotPath = distutils.spawn.find_executable(binName)
            if not dotPath:
                continue
            dotPath = dotPath.rpartition('/')[0]
            if not dotPath in os.environ['PATH']:
                os.environ['PATH'] += os.pathsep + dotPath
            break
    except:
        pass

except ImportError:

    class PlaceholderDiGraph(object):
        """
        Placeholder for graph lib. 
        """
        def __init__(self):
            self.nodes = {}
            self.edges = set()
            
        def __iter__(self):
            for node in self.nodes:
                yield node
        
        def __repr__(self):
            return '<PH Digraph - %d nodes and %d edges>' % (len(self.nodes), len(self.edges))
                    
        def add_node(self, node):
            self.nodes[node] = {}
            
        def remove_node(self, node):
            if node in self.nodes:
                del self.nodes[node]
                self.edges = set(edge for edge in self.edges if not node in edge)
                            
        def add_edge(self, start, end):
            if not start in self.nodes:
                self.add_node(start)
            if not end   in self.nodes:
                self.add_node(end)
                
            self.edges.add((start,end))
            
            self.nodes[start]['edges'] = 1 + self.nodes[start].get('edges', 0)
            self.nodes[end  ]['edges'] = 1 + self.nodes[end  ].get('edges', 0)
            
        def remove_edge(self, start, end):
            try:
                self.edges.remove((start, end))
                
                self.nodes[start]['edges'] = self.nodes[start].get('edges', 0) - 1
                self.nodes[end  ]['edges'] = self.nodes[end  ].get('edges', 0) - 1
    
            except KeyError:
                pass
            
            if start in self.nodes and self.nodes[start].get('edges', 0) <= 0:
                del self.nodes[start]
            if end   in self.nodes and self.nodes[end].get('edges', 0) <= 0:
                del self.nodes[end]

    DiGraph = PlaceholderDiGraph
    
    def draw(*args, **kwargs):
        raise NotImplementedError("To draw update graph, ensure nxpd is installed (NetworkX and Graphviz's Dot)")
                

class GraphModel(UpdateModel):
    
    # Global graph of the calculation
    graph = DiGraph()
    
    def __init__(self, *args, **kwargs):
        # Initialize mixins
        super(GraphModel, self).__init__(*args, **kwargs)
        self.graph.add_node(self)
    
    def _graph_attributes(self):
        """Define the attributes of this node for graphing.
           Override in the subclasses to customize the graph nodes.
        """
        return {}
    
    def draw_graph(self):
        for node in self.graph:
            for key,value in node._graph_attributes().items():
                self.graph.nodes[node][key] = value
        return draw(self.graph)
            
    def subscribe(self, listener):
        super(GraphModel, self).subscribe(listener)
        self.graph.add_edge(self, listener)
    
    def unsubscribe(self, listener):
        super(GraphModel, self).unsubscribe(listener)
        self.graph.remove_edge(self, listener)


from ..transform import Transform


class Unvacuum(Transform):
    """Inject empty groups to fill an index.
    Thus, if the index function is by value, and it's currently
    {1,2,3,6,7,9}
    a delta of 1 starting with 1 would add empty groups for
    {4,5,8}
    """
    raise NotImplementedError


import functools

from .update import UpdateModel
from .graph import GraphModel


class MetaComposable(type):
    
    registered = {}
    
    def __new__(cls, clsname, bases, attrs):
        newclass = super(MetaComposable, cls).__new__(cls, clsname, bases, attrs)
        cls.registered[clsname] = newclass
        return newclass
    
    
class Composable(GraphModel,UpdateModel):
    """All composable classes will scan and consume data
       as well as have a result that can be chained into the next.
    """
    __metaclass__ = MetaComposable
    __slots__ = ('scanners', '_resultSet', '_needsUpdate')
    
    def __init__(self, *args, **kwargs):
        # Initialize mixins
        super(Composable, self).__init__(*args, **kwargs)
        self._needsUpdate = True
        
    def _debounce_scanners(self):
        """As we iterate the scanners, one may exhaust prematurely,
           leaving them in an unaligned state (the first scanners
           may have iterated while the stopped won't).
        """
        # Check if any scanners failed to finish or needs to be reprimed
        if all(scanner.exhausted for scanner in self.scanners):
            return
        for scanner in self.scanners:
            # Once the exhausted scanner is reached, stop. 
            # Only the previous would have over-emitted.
            if scanner.exhausted:
                break
            else:
                scanner.rewind()
                
    def updateFirst(function):
        @functools.wraps(function)
        def ensureUpdated(self, *args, **kwargs):
            if self._needsUpdate:
                self.apply()
            return function(self, *args, **kwargs)
        return ensureUpdated
    
    @updateFirst
    def __iter__(self):
        return (group for group in self.results.groups)

    @updateFirst
    def __getitem__(self, selector):
        return self._resultSet[selector]

    @property
    @updateFirst
    def results(self):
        return self._resultSet   

    # This allows us to better control how the graph is followed
    def _replace_sources(self, newSources):
        for source in set(self._sources).difference(set(newSources)):
            source.unsubscribe(self)
        for source in newSources:
            if isinstance(source, Composable):
                source._resultSet.subscribe(self)
            else:
                source.subscribe(self)
        self._sources = newSources
            
    def update(self, oldSelector, newSelector):
        # (None, None) signals that the data is out of date, 
        #  but there is nothing for dependents to do yet.
        self._needsUpdate = True
        self.notify(None,None)
          
    def apply(self):
        for source in self.sources:
            if isinstance(source, Composable):
                source.apply()
        if self._needsUpdate:
            self._apply()
        self._debounce_scanners()
        
    def _apply(self):
        raise NotImplementedError("The base composable class' apply() must be overridden.")



from itertools import izip as zip
import re

try:
    from com.inductiveautomation.ignition.common import BasicDataset
except ImportError:
    from abc import ABCMeta
    
    class BasicDataset():
        __metaclass__ = ABCMeta
        pass
    

class RecordType(object):
    """Inspired by recipe in Python2 standard library.
    https://docs.python.org/2/library/collections.html#collections.namedtuple

    But doesn't follow the same design for easier overloading.
    """
    __slots__ = ('_tuple')
    _fields = tuple()
    _sanitizedFields = tuple()
    _lookup = {}

    #_reprString = 'Record(%s)' % (', '.join("'%s'=%%r" % f for f in _fields),)

    def __init__(self, values):
        self._tuple = self._cast(values)
        assert len(self._tuple) == len(self._fields), 'Expected %d args, but got %d' % (len(self._fields), len(self._tuple))
            
    def _asdict(self):
        return dict(zip(self._fields, self))
    
    @classmethod
    def getGetter(cls, field):
        ix = cls._lookup[field]
        return lambda myself, ix=ix: myself._tuple[ix]
    
    @classmethod
    def keys(cls):
        return cls._fields

    @property
    def values(self):
        """Returns the tuple.
        >>> R = genRecordType('abc')
        >>> R((1,2,3)).values
        (1, 2, 3)
        """
        return self._tuple

    def _replace(self, **keyValues):
        result = self._cast(map(keyValues.pop, self._fields, self))
        if keyValues: # make sure all got consumed by the map
            raise ValueError('Got unexpected field names: %r' % keyValues.keys())
        self._tuple = result

    def __getitem__(self, key):
        try: # EAFP
            return self._tuple[key]
        except (TypeError,IndexError):
            return self._tuple[self._lookup[key]]

    def __iter__(self):
        """Redirect to the tuple stored when iterating."""
        return iter(self._tuple)

    def __repr__(self):
        'Format the representation string for better printing'
        return repr(self._asdict())
        #return self._reprString % self._tuple
    
    def __getnewargs__(self):
        'Return self as a plain tuple.  Used by copy and pickle.'
        return tuple(self)

    def __getstate__(self):
        'Exclude the OrderedDict from pickling'
        pass
        

def genRecordType(header):
    """Returns something like a namedtuple. 
    Designed to have lightweight instances while having many convenient ways
    to access the data.
    """
    if isinstance(header, BasicDataset):
        rawFields = tuple(h for h in header.getColumnNames())
    else:
        rawFields = tuple(h for h in header)    

    numericFieldPrefix = 'C'
    unsafePattern = re.compile('[^a-zA-Z0-9_]')
    sanitizedFields = [unsafePattern.sub('_', rf) for rf in rawFields]
    for i,field in enumerate(sanitizedFields):
        if field[0].isdigit():
            sanitizedFields[i] = '%s%s' % (numericFieldPrefix, field)

    dupeCheck = set()
    for i,(sf,f) in enumerate(zip(sanitizedFields, rawFields)):
        if sf in dupeCheck:
            n = 1
            while '%s_%d' % (sf,n) in dupeCheck:
                n += 1
            sanitizedFields[i] = '%s_%d' % (sf,n)

        dupeCheck.add(sanitizedFields[i])

    sanitizedFields = tuple(sanitizedFields)

    class Record(RecordType):
        """Generated to match the data stored."""
        _fields = tuple(rf for rf in rawFields)
        _sanitizedFields = tuple(srf for srf in sanitizedFields)
        _lookup = dict(kv for kv 
                       in zip(rawFields + sanitizedFields, range(len(_fields))*2) )

        _reprString = 'Record(%s)' % (', '.join("%s=%%r" % f for f in sanitizedFields),)

    # monkey patch for higher speed access   
    for ix, key in enumerate(sanitizedFields):
        setattr(Record, key, property(lambda self, ix=ix: self._tuple[ix]))
        
    if len(Record._fields) == 1:
        setattr(Record, '_cast', lambda self,v: (v,))
    else:
        setattr(Record, '_cast', lambda self,v: tuple(v))
        
    return Record



from ..calculation import Calculation
from ..scanners.element import ElementScanner


class Sweep(Calculation):
    
    ScanClass = ElementScanner
   
    # Record by record
    def calculate(self):
        """Run the function by row creating a new group.
           If groups don't matter, this is easiest.
           Each resulting group is an update.
        f(a,b)=a+b 
        rs.a = [(1,2,3,4),(5,6),(7,8,9)]
        rs.b = [(0,1,0,1),(0,1),(0,1,0)]
        calc = [(1,3,3,5,5,7,7,9,9)]     # 1 group of 9
        """
        self._resultSet.append(self.function(*values)
                               for values 
                               in zip(*self.scanners))


from ligature.simulators.mixins.support import MixinFunctionSupport
from ligature.expression import Expression


class ExpressionMixin(MixinFunctionSupport):

	# Required overrides for mixin functions
	
	def _configure_default_(self, variable):
		return {}
		
	def _configure_function_(self, expression):
		return Expression(expression)

	# Additional overrides to intercept configuration
	
	def _resolve_variable_definition(self, variable_definition):
		if isinstance(variable_definition, (str, unicode)):
			return {
				'kind': 'Expression',
				'config': {
					'expression': variable_definition
				}
			}
		
		return super(ExpressionMixin, self)._resolve_variable_definition(variable_definition)


	def _resolve_arguments(self, some_callable):
		if isinstance(some_callable, Expression):
			return some_callable._fields
		return super(ExpressionMixin, self)._resolve_arguments(some_callable)
		

	def _initialize_conditional(self, conditional):
		if isinstance(conditional, (str, unicode)):
			return self._close_function(Expression(conditional))
			
		return super(ExpressionMixin, self)._initialize_conditional(conditional)


from ..transform import Transform
from ..recordset import RecordSet
from ..scanners.group import GroupScanner

from itertools import izip as zip


class Pivot(Transform):
    """Rotate groups of records into a record of lists.
    [({a:4,b:3},{a:6,b:5},{a:8,b:7}),({a:10,b:9},{a:12,b:11})]
    becomes
    [({a:(4,6,8),b:(3,5,7)}),({a:(10,12),b:(9,11)})]
    """
    ScanClass = GroupScanner

    def __init__(self, source):
        # Initialize mixins
        super(Pivot, self).__init__(source)

        self.sources = (source,)
        self._resultSet = RecordSet(recordType=source._RecordType)
        self.scanners = (self.ScanClass(self.sources[0]),)
        
    def transform(self):
        for group in self.scanners[0]:
            self._resultSet.append( 
                # cast to the record early so the tuples are not misunderstood
                self._resultSet.coerceRecordType(
                    tuple(zip(*group)) ) )



import unittest


from ligature.recordset import RecordSet
from ligature.examples import simpleRecordSet, simpleAddition

from ligature.calculations.aggregate import Aggregate


class AggregateTestCase(unittest.TestCase):

	def test_basic(self):

		function = lambda a,b: sum(a) - sum(b)

		srs = RecordSet(simpleRecordSet)
		
		c = Aggregate([srs], function, 'c')

		# Calculations are lazily evaluated
		self.assertEqual(c._resultSet._groups, [])

		# When evaluated, we get the following
		self.assertEqual(
			[[v.c for v in group] for group in c.results.groups],
			[[41]]
			)

		srs.extend(simpleAddition)

		# adding data from a source doesn't immediately update
		self.assertEqual(len(c._resultSet._groups), 1)

		# but upon evaluation we see an update has been applied
		# note that results are always one group
		self.assertEqual(
			[[v.c for v in group] for group in c.results.groups],
			[[119]]
			)

		# Demonstrate slicing for columns works as expected
		self.assertEqual(
			[tuple(group) for group in c.results['c',:]],
			[(119,)]
			)

		

def runTests():
	suite = unittest.TestLoader().loadTestsFromTestCase(AggregateTestCase)
	unittest.TextTestRunner(verbosity=2).run(suite)



if __name__ == '__main__':
    unittest.main()


from ..scanner import Scanner 
from .element import ElementScanner
from .chunk import ChunkScanner
from .record import RecordScanner
from .group import GroupScanner


class ReplayingScanner(Scanner):
    """Continues to yield from the same place until ready is called."""
    
    __slots__ = ('_group_anchor', '_record_anchor',)
    
    def __init__(self, *args, **kwargs):
        super(ReplayingScanner, self).__init__(*args, **kwargs)

    def reset(self):
        super(ReplayingScanner, self).reset()
        self.anchor()
        
    def anchor(self):
        """Moves the last acknowledged position to the current."""
        # Cleanup after iteration means this function can see
        #   two types of values: one while iterating and one statically.
        if self._iterating_group and self._iterating_record:
            self._group_anchor = self._group_cursor - 1
        else:
            self._group_anchor = self._group_cursor
        
        self._record_anchor = self._record_cursor

    @property
    def _anchored_group(self):
        return self.source._groups[self._group_anchor]

    @property
    def _anchored_record(self):
        return self._anchored_group[self._record_anchor]
    


class ReplayingElementScanner(ReplayingScanner, ElementScanner):
    """Emits a record's field data until exhausted like a normal generator,
       but if the ratchet isn't set, the next call to the generator
       will simply resume from where the last ratchet was.
    """
    def __iter__(self):
        self._record_cursor = self._record_anchor
        self._group_cursor = self._group_anchor
            
        return super(ReplayingElementScanner,self).__iter__()

    
class ReplayingChunkScanner(ReplayingScanner, ChunkScanner):
    """Emits groups until exhausted like a normal generator,
       but if the ratchet isn't set, the next call will 
       repeat from where the last ratchet was.
    """
    def __iter__(self):
        self._group_cursor = self._group_anchor
        return super(ReplayingChunkScanner,self).__iter__()


class ReplayingRecordScanner(ReplayingScanner, RecordScanner):
    """Emits a record's field data until exhausted like a normal generator,
       but if the ratchet isn't set, the next call to the generator
       will simply resume from where the last ratchet was.
    """
    def __iter__(self):
        self._record_cursor = self._record_anchor
        self._group_cursor = self._group_anchor
            
        return super(ReplayingRecordScanner,self).__iter__()


class ReplayingGroupScanner(ReplayingScanner, GroupScanner):
    """Emits groups until exhausted like a normal generator,
       but if the ratchet isn't set, the next call will 
       repeat from where the last ratchet was.
    """
    def __iter__(self):
        self._group_cursor = self._group_anchor
        return super(ReplayingGroupScanner,self).__iter__()        


import random

class DrunkenWalk(object):
    """A random walker that can get progressively more or less random over time."""    
    
    _reasonable = dict(
        money=100.0,
        tolerance=3.0, # ABV div 3, soberiety gained per recoveryRate steps
        alcoholism=100,
        recoveryRate=1000.0
    )
    
    _maxBounce = 0.2
    _antiTopple = (0.0, 0.0) # (0.2,0.8)
            
    def __init__(self, initValue=0, inebriation=0.12, 
                 money=None, tolerance=None, alcoholism=None, recoveryRate=None,
                 handrails=(0,10.0), leaning=None, stride=None,
                 boozeMenu=[(10,0.12)]):
        """Perform a random walk. The more inebriated, the more the value wanders.
        
        If money is given, more booze will be bought and indebriation gets worse each step.
        Inebriation goes down each step depending on the tolerance.
        
        Alcoholism determines how many steps before another drink it purchased.
        """
        self.value = initValue
        self.inebriation = inebriation
        self.money = money or 0
        self.alcoholism = alcoholism or 0        
        self.tolerance = tolerance or 0
        self.recoveryRate = recoveryRate or 0
        if self.tolerance and not self.recoveryRate:
            self.recoveryRate = self._reasonable['recoveryRate']

        self.leaning = leaning or 0.5
        self.stride = stride or 2
        
        self.boozeMenu = boozeMenu
        self.steps = 0
        self.handrails = handrails
        
        
    def stumble(self):
        
        self.steps += 1
        
        if self.alcoholism and (self.steps % self.alcoholism == 0):
            self.drink()
        if self.tolerance:
            self.inebriation -= self.tolerance / self.recoveryRate
            self.inebriation = max((0, self.inebriation))
        
        self.leaning += (self.leaning * 0.1)
        if self._antiTopple:
            left,right = self._antiTopple
            if self.leaning > right:
                self.leaning = right
            elif self.leaning < left:
                self.leaning = left
        
        self.value += (random.random()-(self.leaning+0.5))*self.inebriation*self.stride
        
        if self.handrails:
            left,right = self.handrails
            if self.value < left:
                self.value = left
                self.leaning = 0.0 + abs(self.leaning/2)
            elif self.value > right:
                self.value = right
                self.leaning = 0.0 - abs(self.leaning/2)
        
        return self.value
    
            
    def traipse(self, steps=None):
        for step in range(steps or self.alcoholism):
            _ = self.stumble()
        return self.value
            
            
    def drink(self):
        if self.money:
            cost,abv = random.choice(abv for cost,abv in self.boozeMenu if cost <=self.money)
            self.money -= cost
            self.inebriation += abv/(self.tolerance)


from transitions.extensions import GraphMachine


class GraphMixin(GraphMachine):
	
	def _init_graphviz_engine(self, use_pygraphviz):
		
		Graph = super(GraphMixin, self)._init_graphviz_engine(use_pygraphviz)
		
		class TweakedGraph(Graph):
		
			_TRANSITION_CHECK = self._TRANSITION_CHECK
			
			def _transition_label(self, tran):
				if tran.get('trigger') == self._TRANSITION_CHECK:
					return ''
				else:
					return super(TweakedGraph, self)._transition_label(tran)
			
		return TweakedGraph





from .compose import Composable
from .scanner import Scanner
from .recordset import RecordSet
from .record import genRecordType
from .expression import Expression


def getArguments(function):
    if isinstance(function, Expression):
        return function._fields
    else:
        return function.__code__.co_varnames[:function.__code__.co_argcount]


class Calculation(Composable):
    """Base class for sweeping over RecordSets.
    """
    __slots__ = ('function', '_mapInputs')
    
    ScanClass = Scanner
    
    def __init__(self, sources, function, outputLabels, mapInputs={}):
        # Initialize mixins
        super(Calculation, self).__init__(sources, function, outputLabels, mapInputs={})
      
        self._resultSet = RecordSet(recordType=genRecordType(outputLabels))
        self.subscribe(self._resultSet)
        self.sources = tuple(sources)

        if isinstance(function, (str,unicode)):
            self.function = Expression(function)
        else:
            self.function = function
            
        self._mapInputs = mapInputs
        self._resolveSources()
    
    def _resolveSources(self):
        """Sources is treated as a stack: last in, first out. 
        Thus if we have sources = (s1,s2), and both have column 'z',
          then we expect to resolve to s2's z
        """
        scanners = []
        columns = [self._mapInputs.get(arg,arg) for arg in getArguments(self.function)]
        for column in columns:
            for source in reversed(self.sources):
                if isinstance(source, Composable):
                    source = source.results
                    
                if column in source._RecordType._lookup:
                    scanners.append(self.ScanClass(source, column))
                    source.subscribe(self)
                    break
                    
            else: # should not complete loop - a column must be found and break!
                raise ValueError('Column "%s" not found in sources!' % column) #' \n\t%s' % (column, '\n\t'.join('{%s}' % s._lookup.keys() for s in reversed(self.sources)))

        self.scanners = tuple(scanners)

    def _apply(self):
        self.calculate()
        self._needsUpdate = False
            
    def _graph_attributes(self):
        label = 'In: %s\\lf(x): "%s"\\lOut: %s' % (
            ', '.join(getArguments(self.function)),
            type(self).__name__,
            ', '.join(self._resultSet._RecordType._fields))
        return {
            'label': label,
            'shape': 'rect'
        }
    
    def calculate(self):
        raise NotImplementedError("The base calculation class' _calculate() must be overridden.")



from ligature.simulators.mixins.support import MixinFunctionSupport
from ligature.easing import Easing


class EasingMixin(MixinFunctionSupport):
	
	# Required overrides for mixin functions

	def _configure_default_(self, variable):
		return {
			'start': self._variables[variable],
			'time_start': self._variables.get(self._escapement_variable, self._DEFAULT_START_VALUE),
		}
	
	def _configure_function_(self, **configuration):
		return Easing(**configuration)



from ..scanner import Scanner


class ElementScanner(Scanner):
    def __iter__(self):
        self._pending_finally()
        for group in self.source._groups[self._group_cursor:]: # error here merely stops iteration
            self._group_cursor += 1
            for record in self._iterGroup(group):
                yield self.getter(record)
        
    def rewind(self, steps=1):
        """Go back the given number of steps in the iteration."""
        while steps > 0:
            steps -= 1
            # If the record cursor is nonzero, back it off by one
            if self._record_cursor > 0:
                # remember, the group cursor aggressively iterates
                if self._group_cursor >= len(self.source._groups):
                    self._group_cursor = len(self.source._groups)-1 
                else:
                    self._record_cursor -= 1
            # If the record cursor is at the start rewind the group
            elif self._group_cursor:
                if self._group_cursor >= len(self.source._groups):
                    self._group_cursor = len(self.source._groups)-1 
                else:
                    self._group_cursor -= 1
                self._record_cursor = len(self.source._groups[self._group_cursor]) -1
            else: # already at the start
                return



from .record import RecordType, genRecordType
from .update import UpdateModel
from .graph import GraphModel

import functools, math
from itertools import izip as zip
from itertools import islice

from weakref import WeakSet

try:
    from com.inductiveautomation.ignition.common import BasicDataset
except ImportError:
    from abc import ABCMeta
    class BasicDataset():
        __metaclass__ = ABCMeta
        pass


class RecordSetColumn(object):
    __slots__ = ('_source', '_index')
    
    def __init__(self, recordSet, columnIndex):
        self._source = recordSet
        self._index = columnIndex

    def __iter__(self):
        """Redirect to the tuple stored when iterating."""
        return ((record._tuple[self._index]
                  for record 
                  in group)
                for group in self._source._groups)
    
    def __getitem__(self, selector):
        """Returns a group or a slice of groups, where the selected column's value in the record is returned.
        """
        if isinstance(selector, slice):
            if selector.step:
                raise NotImplementedError("Columns should not be sliced by steps. Window the RecordSet and group records instead.")
            
            return ((record._tuple[self._index]
                      for record
                      in group )
                    for group 
                    in islice(self._source._groups, selector.start, selector.stop) )
        else:
            return (record._tuple[self._index]
                      for record
                      in self._source._groups[selector] )

    def __repr__(self):
        'Format the representation string for better printing'
        return 'RecordSetColumn("%s" at %d)' % (self._source._RecordType._fields[self._index], self._index)

        

class RecordSet(GraphModel,UpdateModel):
    """Holds groups of records. The gindex is the label for each of the tuples of Records.
    
    Based on collections.MutableSequence
    """
    # Track all the RecordSets that get made. We can use this to automagically trim the cache.
    # References need to be weak to ensure garbage collection can continue like normal.
    _instances = WeakSet()

    __slots__ = ('_RecordType', '_groups', '_columns')


    def __new__(cls, *args, **kwargs):
        """See https://stackoverflow.com/a/12102666/1943640 for an example of this."""
        instance = super(RecordSet, cls).__new__(cls, *args, **kwargs)
        cls._instances.add(instance)
        return instance

    @classmethod
    def _truncateAll(cls):
        instances = list(cls._instances)
        for instance in instances:
            instance.truncate()

    def truncate(self):
        """Clear out data that is not unused.
           Cooperate with the scanners pointing to this RecordSet
             by asking for a tuple declaring what records or groups
             it needs.
           Only groups are truncated, and only from the beginning.
           Once completed, each listening scanner is notified
             so its cursors can be corrected accordingly.
        """
        raise NotImplementedError

        safeGroupIx = -1
        for scanner in listeningScanners:
            pass

        for scanner in listeningScanners:
            scanner.updateCursorsForRemoval(safeGroupIx)


    # INIT
    def _initializeDataSet(self, dataset, validate=False):
        """Convert the DataSet type into a RecordSet
        """
        self._RecordType = genRecordType(dataset.getColumnNames())
        columnIxs = range(len(self._RecordType._fields))
        records = []
        for rix in range(dataset.getRowCount()):
            row = tuple(dataset.getValueAt(rix, cix) for cix in columnIxs)
            records.append(self._RecordType(row))
        self._groups = [tuple(records)]
        
    def _initializeEmpty(self, RecordType):
        """Simply define what kind of RecordSet this will be, but start with no data.
        """
        self._RecordType = RecordType
        self._groups = []
    
    def _initializeRaw(self, RecordType, data):
        """Make a list with a single tuple entry, that was the list of new records.
           Using a generator as the tuple argument is about 4-10x slower.
        """
        self._RecordType = RecordType
        self._groups = [tuple([RecordType(record) 
                               for record 
                               in data])]
    
    def _initializeRecords(self, records, validate=False):
        """Initialize RecordSet from the records provided.
        Optionally validate the list to ensure they're all the same kind of record.
        """
        assert len(records) > 0, """If only records are provided, an example entry must be included: `len(records) == 0`"""
        self._RecordType = type(records[0])
        if validate:
            assert all(isinstance(r, RecordType) for r in records), 'All entries were not the same RecordType'
        self._groups = [tuple(records)]

    def _initializeCopy(self, recordSet):
        self._RecordType = recordSet._RecordType
        self._groups = [group for group in recordSet._groups]
        # Note that it'll regenerate indexes even on copy...
        
    
    def __init__(self, initialData=None,  recordType=None, initialLabel=None, validate=False):#, indexingFunction=None):        
        """When creating a new RecordSet, the key is to provide an unambiguous RecordType,
             or at least enough information to define one.
        """
        # Initialize mixins
        super(RecordSet, self).__init__(initialData, recordType, initialLabel, validate)
        
        # We can initialize with a record type, a record, or an iterable of records
        # First check if it's a DataSet object. If so, convert it.
        if isinstance(initialData, BasicDataset):
            self._initializeDataSet(initialData)
        elif recordType:
            # create a RecordType, if needed
            if not (isinstance(recordType, type) and issubclass(recordType, RecordType)):
                recordType = genRecordType(recordType)
            if initialData:
                self._initializeRaw(recordType, initialData)
            else:
                self._initializeEmpty(recordType)
        elif initialData:
            if isinstance(initialData, RecordSet):
                self._initializeCopy(initialData)
            else:
                self._initializeRecords(initialData, validate)
        else:
            raise ValueError("""Insufficient information to initialize the RecordSet."""
                             """ A RecordType must be implied by the constructor arguments.""")
                
        # monkey patch for higher speed access
        self._columns = tuple(RecordSetColumn(self, ix) 
                              for ix 
                              in range(len(self._RecordType._fields)))
            
            
    def clear(self):
        self._groups = []
        
        
    def column(self,column):
        return self._columns[self._RecordType._lookup[column]]

    
    def coerceRecordType(self, record):
        return self._RecordType(record)

    # Sized
    def __len__(self):
        """Not terribly useful - this only tells how many chunks there are in the RecordSet.
           Use this as a poor man's fast cardinality check against other RecordSets.
        """
        return len(self._groups)
    
    # Iterable 
    # Sequence
    
    def __getitem__(self, selector):
        """Get a particular set of records given the selector.
           Note that this is record-centric, like the iterator.
        """
        if isinstance(selector, tuple):
            column,slicer = selector
            return self.column(column)[slicer]
        elif isinstance(selector, slice):
            return islice(self.records, selector)     
        elif isinstance(selector, int):
            index = selector
            for group in self._groups:
                if len(group) > index:
                    return group[index]
                index -= len(group)
            else:
                raise IndexError("There are not enough records in the groups to meet the index %d" % index)
        else:
            raise NotImplementedError("The selector '%r' is not implemented" % selector)
            #return self._groups[selector]
    
    @property
    def groups(self):
        return (group 
                for group 
                in self._groups)
    
    @property
    def records(self):
        return (record 
                for group in self._groups 
                for record in group)
    
    def __iter__(self):
        return self.records
        #return (recordGroup for recordGroup in self._groups)

    # Container
    def __contains__(self, search):
        """Search from the start for the search object.
           If a record is provided, then the groups will themselves be 
             exhaustively searched.
        """
        if isinstance(search, self._RecordType):
            for group in self._groups:
                if search in group:
                    return True
        elif search in self._groups:
            return True
        return False

    def __reversed__(self):
        """Reverses BOTH the groups and the records in the groups.
           Pure generators are returned.
        """
        return (reversed(group) for group in reversed(self._groups))
                
    def index(self, search, fromTop=True):
        """Returns the index of the group the search item is found in."""
        if fromTop:
            iterDir = (gg for gg in reversed(enumerate(self._groups)))
        else:
            iterDir = (gg for gg in enumerate(self._groups))
            
        if isinstance(search, self._RecordType):
            for gix, group in iterDir:
                if record in group:
                    return gix
        elif isinstance(search, tuple):
            for gix, group in iterDir:
                if search == group:
                    return gix
                for record in group:
                    if record._tuple == search:
                        return gix
        raise ValueError('Search item %r was not found' % search)

    def count(self, search):
        """Returns the number of times the search item is found.
           Note that this tests for equivalency, not instantiation!
        """
        if isinstance(search, self._RecordType):
            return sum(1
                       for group in self._groups
                       for record in group
                       if record == search )
        elif isinstance(search, tuple):
            return sum(1 for group in self._groups if search == group)
        return 0

    def append(self, addition):
        """Append the group of records to the end of _groups.
           If a record is given, an group of one will be appended.
        """
        if isinstance(addition, RecordSet):
            self.extend(addition)
        else:
            if isinstance(addition, self._RecordType):
                newGroup = (self.coerceRecordType(addition),)
            else: 
                # tuple creation is slightly faster if the generator is consumed by a list first
                newGroup = tuple([self.coerceRecordType(entry)
                                  for entry
                                  in addition ])
            self._groups.append(newGroup)
            self.notify(None, -1)
    
    def extend(self, additionalGroups):
        """Extend the records by concatenating the record groups of another RecordSet.
        """
        if isinstance(additionalGroups, RecordSet):
            assert self._RecordType._fields == additionalGroups._RecordType._fields, 'RecordSets can only be extended by other RecordSets of the same RecordType.'
            self._groups.extend(additionalGroups._groups)
            self.notify(None, slice(-len(additionalGroups),None))
        else:
            for group in additionalGroups:
                self.append(group)
    
    def __iadd__(self, addition):
        """Overload the shorthand += for convenience."""
        if isinstance(addition, self._RecordType):
            self.append(addition)
        else:
            self.extend(addition)
        return self
    
    def _graph_attributes(self):
        fields = ', '.join(self._RecordType._fields)
        label = 'RecordSet\n%s' % fields
        return {
            'label': label,
            'shape': 'cylinder',
        }
    
    def __str__(self):
        return 'RecordSet=%r' % repr(self._RecordType._fields)
               
    def __repr__(self, elideLimit=20):
        'Format the representation string for better printing'
        records = list(islice((r for r in self.records), elideLimit))
        totalRecordCount = sum(len(g) for g in self.groups)
        out = ['RecordSet with %d groups of %d records' % (len(self), totalRecordCount)]
        # preprocess
        maxWidths = [max([len(f)] + [len(repr(v))+1 for v in column]) 
                     for f,column 
                     in zip(self._RecordType._fields,
                            zip(*records) 
                               if records 
                               else ['']*len(self._RecordType._fields))]
        digits = math.floor(math.log10(elideLimit or 1)) + 1
        maxGixWidth = math.floor(math.log10(len(self._groups) or 1)) + 1
        prefixPattern = ' %%%ds  %%%ds |' % (maxGixWidth, digits)
        recordPattern = prefixPattern + ''.join(' %%%ds' % mw for mw in maxWidths)
        out += [recordPattern % tuple(['',''] + list(self._RecordType._fields))]
        out += [recordPattern % tuple(['',''] + ['-'*len(field) for field in self._RecordType._fields])]
        
        remaining = elideLimit
        for gix,group in enumerate(self._groups):
            for j,record in enumerate(group):
                out += [recordPattern % tuple(['' if j else gix,j] + [repr(v) for v in record])]
                remaining -= 1
                if not remaining:
                    break
            if not remaining:
                break
        if totalRecordCount > elideLimit:
            out += ['  ... and %d elided' % (totalRecordCount - elideLimit)]   
        out += ['']

        return '\n'.join(out)




from ..calculation import Calculation
from ..scanners.element import ElementScanner


class Aggregate(Calculation):
    
    ScanClass = ElementScanner
    
    def calculate(self):
        """Run the aggregate function by group, each creating a new group.
           Useful for aggregates that may be used with another calc's groups.
           Collapses down a group to a single value.
        f(a,b)=sum(a)-sum(b) 
        rs.a = [(1,2,3,4),(5,6),(7,8,9)]  = 45
        rs.b = [(0,1,0,1),(0,1),(0,1,0)]  = 4
        calc = [(41,)]                    # 1 group of 1
        """
        for scanner in self.scanners:
            scanner.reset()
        
        self._resultSet.clear()
        
        self._resultSet.append( (self.function(*self.scanners),) )


"""
    Convert strings safely into properly executable functions (not just compiled!)
"""

import __builtin__
import operator as op
import tokenize
from ast import literal_eval
from StringIO import StringIO

from .enum import MetaEnum, Enum


__copyright__ = """Copyright (C) 2020 Corso Systems"""
__license__ = 'Apache 2.0'
__maintainer__ = 'Andrew Geiger'
__email__ = 'andrew.geiger@corsosystems.com'

__all__ = ['Expression']

TOKENS = MetaEnum(
     'TOKENS', 
     (Enum,), 
     dict((attr,getattr(tokenize,attr)) 
          for attr 
          in dir(tokenize) 
          if    attr.upper() == attr 
            and isinstance(getattr(tokenize, attr),int)))


tokenTypeLookup = dict(
    (getattr(tokenize, tokenType), getattr(TOKENS, tokenType))
    for tokenType
    in dir(tokenize)
    if isinstance(getattr(tokenize, tokenType), int)
       and tokenType.upper() == tokenType
)


def overload_concat_add(*args):
    if len(args) == 1:
        return op.pos(args[0])
    if all(hasattr(type(v),'__iter__') for v in (args)):
        return op.concat(*args)
    else:
        return op.add(*args)

def overload_neg_sub(*args):
    if len(args) == 1:
        return op.neg(args[0])
    else:
        return op.sub(*args)

def overload_mul_rep(*args):
    if hasattr(type(args[0]),'__iter__') and isinstance(args[1],int):
        return op.repeat(*args)
    else:
        return op.mul(*args)

def extend_iterable(left,right):
    if isinstance(left, tuple):
        return left + (right,)
    else:
        return (left,right)


one_argument_operators = {
    'not': op.not_, '!': op.not_,
}

two_argument_operators = {
    '+' : overload_concat_add,
    'in': op.contains,
    '/' : op.truediv, # see PEP 238
    '//': op.floordiv,
    '&' : op.and_, 'and': op.and_,
    '^' : op.xor,
    '~' : op.invert,
    '|' : op.or_, 'or': op.or_,
    '**': op.pow,
    'is': op.is_,
    #'is not': op.is_not,
    '<<': op.lshift,
    '%' : op.mod,
    '*' : overload_mul_rep, #op.mul,
    '-' : overload_neg_sub,
    '>>': op.rshift,
    '<' : op.lt,
    '<=': op.le,
    '==': op.eq,
    '!=': op.ne,
    '>=': op.ge,
    '>' : op.gt,
    ',': extend_iterable,
#   '.': getattr,

     # backwards because we're hacking a parenthetical replacement, and it's like '.'
    '__getitem__': lambda attr, obj: obj.__getitem__(attr),
}


precedence = {
    20: ['(',')','[',']'],
    0: [','],
    1:  ['lambda'],
    #2:  ['if else'],
    3:  ['or'],
    4:  ['and'],
    5:  ['!','not'],
    6:  ['in', 'is', '<', '<=', '>', '>=', '<>', '!=', '=='], # 'not in', 'is not', 
    7:  ['|'],
    8:  ['^'],
    9:  ['&'],
    10: ['<<','>>'],
    11: ['+','-'],
    12: ['*','/','//','%'],
    13: ['~'], # ['+','-'] # bitwise not, pos, neg
    14: ['**'],
#   18: ['.'],
    18: ['__getitem__']
    #15: [index, slices, x(call), x.attribute]
    #16: [(tuple), [list], {key: value},
}

precedenceLookup = dict(
    (token, key)
    for key,tokens in precedence.items()
    for token in tokens
)



def gather_tokens(expression):
    return (
        (tokenTypeLookup[tokenType], token)
        for tokenType, token, (srow,scol), (erow,ecol), line 
        in tokenize.generate_tokens(StringIO(expression).readline)
    )


def convert_to_postfix(expression):
    opstack = []
    output = []

    raw_tokens = gather_tokens(expression.strip())

    # preprocess for lookups and patchwerks
    tokens = []
    for tokenType, token in raw_tokens:
        if tokenType == TOKENS.OP and token == '[':
            #tokens.append((TOKENS.OP, '.'))
            tokens.append((TOKENS.OP, '__getitem__'))
            tokens.append((TOKENS.OP, '('))
        elif tokenType == TOKENS.OP and token == ']':
            tokens.append((TOKENS.OP, ')'))
        
        elif tokenType == TOKENS.STRING:
            tokens.append((TOKENS.STRING, token[1:-1]))
        
        elif tokenType == TOKENS.ERRORTOKEN and token.strip() == '':
            pass
        
        # ensure the word-like tokens are correctly identified
        elif tokenType == TOKENS.NAME and token in one_argument_operators:
            tokens.append((TOKENS.OP, token))
        
        elif tokenType == TOKENS.NAME and token in two_argument_operators:
            tokens.append((TOKENS.OP, token))
        
        else:
            tokens.append((tokenType, token))

    # Handle the tokens gathered in order.
    # Assume that tokens are provided in INFIX notation
    for tokenTuple in tokens:
        tokenType, token = tokenTuple
        
        # Stop doing work once we get to the end.
        # Importantly: this works for just one line!
        if tokenType == tokenize.ENDMARKER: break
            
        if tokenType == tokenize.OP:
            # Closing a group is treated a bit specially, since it can partly drain the opstack
            if token == ')':
                
                # Drain the stack until we get to a closing parenthesis 
                #   or the calling function name 
                #   (names only appear on the stack when immediately followed by a parens) 
                while opstack and not (   (opstack[-1] == (tokenize.OP,'('))
                                       or opstack[-1][0] == tokenize.NAME):
                    output.append(opstack.pop())
                # If the parens was preceded by a call, now move that to the output
                if opstack and opstack[-1][0] == tokenize.NAME:

                    output.append(opstack.pop())
                    
                    # Count the compounding
                    dots = 0
                    while opstack and opstack[-1][0] == tokenize.NAME:
                        dots += 1
                        output.append(opstack.pop())
                    for i in range(dots):
                        if opstack and opstack[-1][0] == tokenize.OP and opstack[-1][1] == '.':
                            output.append(opstack.pop())
                        else:
                            raise AttributeError("Unexpected token: was expecting the '.' operator, but the stack is this instead\nopstack:%s\noutput:%s" % (
                                str(opstack), str(output)))
                        
                # ... otherwise it must just be the opening parens operator
                else:
                    _ = opstack.pop()
                    
            # Otherwise the token is just a part of the formula 
            else:
                # If we're starting a parenthetical group, check if there's a name right before it.
                #   If so, then assume it's a call, and add that to the stack instead of the '('
                if token == '(' and output and (output[-1][0] == tokenize.NAME or output[-1] == (tokenize.OP, '.')):
                    if output[-1] == (tokenize.OP, '.'):
                        opstack.append(output.pop())    
                        opstack.append(output.pop())
                        opstack.append((TOKENS.OP, '('))
                    
                    # Check if this is a compound attribute, if so pop an extra off
                    elif opstack and opstack[-1][0] == tokenize.OP and opstack[-1][1] == '.':
                        opstack.append(output.pop())
                        if output and output[-1][0] == tokenize.NAME:
                            opstack.append(output.pop())
                        else:
                            raise NotImplementedError('Attributes in expressions must be on named tokens')
                    else:
                        opstack.append(output.pop())
                                    
                elif token == '.':
                    opstack.append(tokenTuple)
                
                # Otherwise it's a normal token that has to follow the rules of precedence
                else:
                    # get the value of this token in relation to others
                    tokenPrecedence = precedenceLookup[token]
                    # ... and drain the opstack of things that should come first
                    while (opstack 
                           and precedenceLookup.get(opstack[-1][1],-20) >= tokenPrecedence 
                           and not (opstack[-1][0] == tokenize.OP
                                    and opstack[-1][1] in ('(','[','{'))):
                        opToken = opstack.pop()
                        output.append(opToken)
                    # ... and once we know this is the highest precedence operation, add it to the stack
                    opstack.append(tokenTuple)

        # All non-operators get pushed to the stack.
        # These are normal things like numbers and names
        else:
            output.append(tokenTuple)
            
            # check if this gets resolved as an attribute (effectively very high precedence operator, next to () )
            if opstack and opstack[-1][1] == '.':
                output.append(opstack.pop())
            

        # print '=> %s\n   %-50s\n   %s\n' % (tokenTuple, '   OPS >> %r' % opstack, '<< OUT    %r' % output)

    while opstack:
        output.append(opstack.pop())
    
    return tuple(output)


def isCallable(obj):
    try:
        return bool(obj.__call__)
    except AttributeError:
        return False

        
whitelisted_modules = set((
    'shared',
    'math'
))


whitelisted_builtins = set((
    'max','min'
))


class REF_TYPE(Enum):
    CONSTANT = -2
    ARGUMENT = -4
    FUNCTION = -8
    EXTERNAL = -16


class Expression(object):
    
    __slots__ = ('_fields', '_eval_func',
                 '_arguments', '_constants', '_functions', '_externals'
                )
    
    def __init__(self, expression):
        if isinstance(expression, str):
            # convert the expression to something we can resolve reliably
            postfixStack = convert_to_postfix(expression)
            # ... and map it to the properties here
            self._resolve_function(postfixStack)
        else:
            self._resolve_function(expression)
        
    def _resolve_function(self, postfixStack):
        
        self._arguments = []
        self._constants = []
        self._functions = []
        self._externals = []
        opstack = []
    
        references = {
            REF_TYPE.FUNCTION: self._functions,
            REF_TYPE.ARGUMENT: self._arguments,
            REF_TYPE.CONSTANT: self._constants,
            REF_TYPE.EXTERNAL: self._externals,
        }
        
        reference_names = {
            REF_TYPE.FUNCTION: 'func',
            REF_TYPE.ARGUMENT: 'args',
            REF_TYPE.CONSTANT: 'const',
            REF_TYPE.EXTERNAL: 'ext',
        }
        
    
        for tokenType,token in postfixStack:

            if tokenType == tokenize.OP:

                if token == '.':
                    #raise NotImplementedError('Better to preprocess and resolve first')
                    
                    # though not strictly necessary since the dot will work as a normal operator, 
                    #   this will resolve down and optimize a bit
                    (argType2,argIx2), (argType1,argIx1) = opstack.pop(), opstack.pop()
                    
                    if argType1 == REF_TYPE.FUNCTION and self._functions[argIx1] in self._externals and argType2 == REF_TYPE.ARGUMENT:
                        
                        attribute = self._arguments.pop(argIx2)
                        
                        external = getattr(self._functions[argIx1], attribute)
                        self._externals.append(external)
                        
                        if isCallable(external):
                            argType3,argIx3 = opstack.pop()
                            argRef3 = references[argType3]
                            
                            if argType3 == REF_TYPE.FUNCTION:
                                self._functions.append(lambda function=external, ar1=argRef3, aix1=argIx3: function(
                                                    ar1[aix1]()
                                                ) )
                            else:
                                self._functions.append(lambda function=external, ar1=argRef3, aix1=argIx3: function(
                                                    ar1[aix1]
                                                ) )

                            opstack.append( (REF_TYPE.FUNCTION, len(self._functions) - 1) )
                        else:
                            self._constants.append(external)
                            opstack.append( (REF_TYPE.CONSTANT, len(self._constants) - 1) )
                        #self._externals.append(getattr(self._externals[argIx1], attribute))
                        #opstack.append( (REF_TYPE.EXTERNAL, len(self._externals) - 1) )
                    
                    # This is something like a.b where a is an argument and b is some sort of attribute of that type.
                    elif (argType1 == REF_TYPE.ARGUMENT or argType1 == REF_TYPE.FUNCTION) and argType2 == REF_TYPE.ARGUMENT:
                        
                        argRefs = references[argType1]
    
                        attribute = self._arguments.pop(argIx2)
                        
                        # Treat as an attribute lookup on the first
                        # Check if it needs calling before resolution, though
                        if argType1 == REF_TYPE.ARGUMENT:
                            self._functions.append(lambda self=self, argRefs=argRefs, aix1=argIx1, attribute=attribute: getattr(
                                                    argRefs[aix1],
                                                    attribute
                                                ) )
                        else: # argType1 == REF_TYPE.FUNCTION:
                            self._functions.append(lambda self=self, argRefs=argRefs, aix1=argIx1, attribute=attribute: getattr(
                                                    argRefs[aix1](),
                                                    attribute
                                                ) )

                        opstack.append( (REF_TYPE.FUNCTION, len(self._functions) - 1) )
                
                    else:
                        raise AttributeError('Not sure what to do with this:\nArg 1: %r at %s as "%r"\nArg 2: %r at %s as "%r"' % (
                                                    argType1,argIx1,references[argType1][argIx1],
                                                    argType2,argIx2,references[argType2][argIx2],))
                    
                
                elif token in two_argument_operators: 
                    (argType2,argIx2), (argType1,argIx1) = opstack.pop(), opstack.pop()
                    
                    argRef1 = references[argType1]
                    argRef2 = references[argType2]
                    
                    function = two_argument_operators[token]
                    
                    # Resolve the way we call the arguments in
                    #   it needs a late binding, so it kind of has to be broken out like this 
                    #   (I don't know how to do this without adding more indirection)
                    if argType1 == REF_TYPE.FUNCTION:
                        if argType2 == REF_TYPE.FUNCTION:
                            self._functions.append(lambda self=self, function=function, ar1=argRef1, aix1=argIx1, ar2=argRef2, aix2=argIx2: function(
                                                ar1[aix1](),
                                                ar2[aix2]()
                                            ) )
                        else:
                            self._functions.append(lambda self=self, function=function, ar1=argRef1, aix1=argIx1, ar2=argRef2, aix2=argIx2: function(
                                                ar1[aix1](),
                                                ar2[aix2]
                                            ) )
                    else:
                        if argType2 == REF_TYPE.FUNCTION:
                            self._functions.append(lambda self=self, function=function, ar1=argRef1, aix1=argIx1, ar2=argRef2, aix2=argIx2: function(
                                                ar1[aix1],
                                                ar2[aix2]()
                                            ) )
                        else:
                            self._functions.append(lambda self=self, function=function, ar1=argRef1, aix1=argIx1, ar2=argRef2, aix2=argIx2: function(
                                                ar1[aix1],
                                                ar2[aix2]
                                            ) )
                            
                    opstack.append( (REF_TYPE.FUNCTION, len(self._functions) - 1) )

                if token in one_argument_operators:
                
                    (argType1,argIx1)= opstack.pop()
                    
                    argRef1 = references[argType1]
                    
                    if argType1 == REF_TYPE.FUNCTION:
                        self._functions.append(lambda self=self, function=function, ar1=argRef1, aix1=argIx1: function(
                                            ar1[aix1]()
                                        ) )
                    else:
                        self._functions.append(lambda self=self, function=function, ar1=argRef1, aix1=argIx1: function(
                                            ar1[aix1]
                                        ) )
                        
                    opstack.append( (REF_TYPE.FUNCTION, len(self._functions) - 1) )
                    
            elif tokenType == tokenize.NAME:
                # Check if it's a variable or module we trust
                # as we encounter the '.' operator, whitelist
                if token in whitelisted_modules:
                    # math.sin(x) --> ((1, 'math'), (1, 'sin'), (51, '.'), (1, 'x'))
                    self._externals.append(__import__(token))
                    self._functions.append(self._externals[-1])
                    opstack.append( (REF_TYPE.FUNCTION, len(self._functions) - 1) )
                    #opstack.append( (EXTERNAL, len(self._externals) - 1) )
                
                elif token in whitelisted_builtins:
                    
                    function = getattr(__builtin__,token)

                    self._externals.append(function)
                    #self._functions.append(self._externals[-1])
                    #opstack.append( (FUNCTION, len(self._functions) - 1) )

                    (argType1,argIx1)= opstack.pop()
                    
                    argRef1 = references[argType1]
                    
                    if argType1 == REF_TYPE.FUNCTION:
                        self._functions.append(lambda self=self, function=function, ar1=argRef1, aix1=argIx1: function(
                                            ar1[aix1]()
                                        ) )
                    else:
                        self._functions.append(lambda self=self, function=function, ar1=argRef1, aix1=argIx1: function(
                                            ar1[aix1]
                                        ) )
                        
                    opstack.append( (REF_TYPE.FUNCTION, len(self._functions) - 1) )                    
                    
                else:
                    if not token in self._arguments:
                        self._arguments.append(token)
                        opstack.append( (REF_TYPE.ARGUMENT, len(self._arguments) - 1) )

            elif tokenType == tokenize.NUMBER:
                self._constants.append(literal_eval(token))
                opstack.append( (REF_TYPE.CONSTANT, len(self._constants) - 1) )

            elif tokenType == tokenize.STRING:
                self._constants.append(str(token))
                opstack.append( (REF_TYPE.CONSTANT, len(self._constants) - 1) )

            # print 'Token: %s, %s' % (tokenTypeLookup[tokenType], token)
            # print '  Opstack: ',['%s[%d]=%s' % (reference_names[t],ix,references[t][ix]) for t,ix in opstack]
            
        self._fields = tuple(self._arguments)
        self._arguments[:] = []
        
        opType,opIx = opstack.pop()
        if opType in (REF_TYPE.CONSTANT, REF_TYPE.ARGUMENT):
            self._eval_func = lambda: references[opType][opIx]
        else:
            self._eval_func = references[opType][opIx]


    def __call__(self, *args, **kwargs):
        if kwargs:
            self._arguments[:] = [kwargs.get(field) or args[i] for i,field  in enumerate(self._fields)]
        else:
            self._arguments[:] = args
        
        if len(self._arguments) < len(self._fields):
            raise TypeError('Expression takes exactly %d argument%s: %r (%d given)' % (
                                len(self._fields), 's' if len(self._fields) > 1 else '', list(self._fields), len(self._arguments)))     
        return self._eval_func()    



from ..scanner import Scanner


class ChunkScanner(Scanner):
    """For a field in a source, this emits values in chunks
       - one chunk for each group in the source.
    """
    def __iter__(self):
        self._pending_finally()
        for group in self.source._groups[self._group_cursor:]: # error here merely stops iteration
            self._group_cursor += 1
            yield tuple(self.getter(record) for record in group)
            
    def rewind(self, steps=1):
        """Go back the given number of steps in the iteration."""
        while steps > 0:
            steps -= 1
            # If we have a nonzero group cursor, then just decrement it
            if self._group_cursor > 0:
                self._group_cursor -= 1
            else: # already at the start
                return



from .recordset import RecordSet


def genData(columns, rows, start=0):
    if not isinstance(columns, int):
        columns = len(columns)
    if columns == 1:
        return (i for i in range(start, start+rows*columns, columns))
    else:
        return (tuple(range(i,i+columns))
                for i 
                in range(start, start+rows*columns, columns))



a1 = [(1,2,3,4),(5,6),(7,8,9)]
b1 = [(0,1,0,1),(0,1),(0,1,0)]

a2 = [(11,12,13),(14,15,16)]
b2 = [(1,0,1),(0,1,0)]

simpleRecordSet = RecordSet(recordType='ab')
for g in zip(a1,b1):
    simpleRecordSet.append(v for v in zip(*g))

simpleAddition = RecordSet(recordType='ab')
for g in zip(a2,b2):
    simpleAddition.append(v for v in zip(*g)) 